#!/usr/bin/env python

# third party imports
import numpy as np
import scipy.interpolate as spint
import logging

from esi_utils_time.ancient_time import HistoricTime
from esi_utils_rupture.base import Rupture
from esi_utils_rupture import constants


class PointRupture(Rupture):
    """
    Rupture class for point sources. The purpose is to gracefully handle:

        - Requests for rupture distances when no rupture is available.
        - Provide reasonable default values for rupture parameters.
    """

    def __init__(self, origin, reference="Origin"):
        """
        Constructs a PointRupture instance.

        Args:
            origin (Origin): Reference to a ShakeMap Origin instance.
            reference (str): Citable reference for rupture; in the case of a
                PointRupture, the 'reference' is probably the origin.

        Returns:
            PointRupture instance.
        """
        self._origin = origin
        self._reference = reference

        coords = [origin.lon, origin.lat, origin.depth]

        d = {
            "type": "FeatureCollection",
            "metadata": {"reference": reference},
            "features": [
                {
                    "type": "Feature",
                    "properties": {"rupture type": "rupture extent"},
                    "geometry": {"type": "Point", "coordinates": coords},
                }
            ],
        }

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, HistoricTime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v

        self._geojson = d
        #
        # Use Wells and Coppersmith (1994) to compute some basic
        # fault parameter based on the magnitude. Use the "All"
        # fault type since if this is a point rupture we probably
        # don't know much.
        #
        width = -1.01 + 0.32 * origin.mag
        self.faultWidth = np.power(10.0, width)
        area = -3.49 + 0.91 * origin.mag
        self.faultArea = np.power(10.0, area)
        self.faultLength = self.faultArea / self.faultWidth

    def getLength(self):
        """
        Return the W&C value based on magnitude.
        """
        return self.faultLength

    def getWidth(self):
        """
        Rupture width.
        Return the W&C value based on magnitude
        """
        return self.faultWidth

    def getArea(self):
        """
        Rupture area
        Return the W&C value based on magnitude
        """
        return self.faultArea

    def getStrike(self):
        """
        Strike, which is None.
        Could potentially get from strec or something?
        """
        return constants.DEFAULT_STRIKE

    def getDip(self):
        """
        Dip, which is None.
        Could potentially get from strec or something?
        """
        return constants.DEFAULT_DIP

    def getDepthToTop(self):
        """
        Depth to top of rupture.
        Use the Kaklamanos et al. (2011) formula:
            ztor = max((Zhyp - 0.6W * sin(delta)), 0)
        with the width coming from W&C 1994 as above.
        The default dip is 90, so we're reduced to:
        """
        ztor = min(30.0, max(self._origin.depth - 0.6 * self.faultWidth, 0))
        return ztor

    def getQuadrilaterals(self):
        return None

    @property
    def lats(self):
        """
        Returns rupture latitudes, which is just the hypocenter for a
        PointRupture."""
        return self._origin.lat

    @property
    def lons(self):
        """
        Returns rupture longitudes, which is just the hypocenter for a
        PointRupture."""
        return self._origin.lon

    @property
    def depths(self):
        """
        Returns rupture depths, which is just the hypocenter for a
        PointRupture."""
        return self._origin.depth

    def computeRjb(self, lon, lat, depth):
        """
        Return epicentral distances.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            array: an array of epicentral distances.
        """
        return self.computeRepi(lon, lat, depth)

    def computeRrup(self, lon, lat, depth):
        """
        Return hypocentral distances.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            array: and array of hypocentral distances
        """
        return self.computeRhyp(lon, lat, depth)

    def computeGC2(self, lon, lat, depth):
        """
        Method for computing version 2 of the Generalized Coordinate system
        (GC2) by Spudich and Chiou OFR 2015-1028.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            dict: Dictionary with keys for each of the GC2-related distances,
                which include 'rx', 'ry', 'ry0', 'U', 'T'.
        """
        # This just returns defaults of zero, which will hopefully behave
        # gracefully as used in GMPEs but should eventually be updated
        # so that things like the hangingwall terms are unbiased.
        gc2 = {
            "rx": np.zeros_like(lon),
            "ry": np.zeros_like(lon),
            "ry0": np.zeros_like(lon),
            "U": np.zeros_like(lon),
            "T": np.zeros_like(lon),
        }
        return gc2



#!/usr/bin/env python3


# Depth tolerance in km (for determining if top and bottom edges are
# horizontal)
DEPTH_TOL = 0.05

# Maximum ratio of distance off of the plane (relative to edge length) for the
# 4th point to be before being considered non-co-planar and adjusted to
# actually be on the plane?
OFFPLANE_TOLERANCE = 0.05

RAKEDICT = {'SS': 0.0, 'NM': -90.0, 'RS': 90.0, 'ALL': None}

DEFAULT_MECH = 'ALL'
DEFAULT_STRIKE = 0.0
DEFAULT_DIP = 90.0
DEFAULT_RAKE = 0.0
DEFAULT_WIDTH = 0.0
DEFAULT_ZTOR = 0.0

ORIGIN_REQUIRED_KEYS = ['id', 'netid', 'network', 'lat', 'lon', 'depth',
                        'locstring', 'mag', 'time']
# Times can have either integer or floating point (preferred) seconds
TIMEFMT = '%Y-%m-%dT%H:%M:%S.%fZ'
ALT_TIMEFMT = '%Y-%m-%dT%H:%M:%SZ'



#!/usr/bin/env python

# stdlib modules
import copy
from datetime import datetime

# third party imports
import numpy as np

try:
    _ = np.RankWarning  # will work on numpy < 2
except AttributeError:
    setattr(np, "RankWarning", RuntimeWarning)  # will work on numpy > 2
from openquake.hazardlib.geo.mesh import Mesh
from openquake.hazardlib.geo.geodetic import point_at
from openquake.hazardlib.geo.point import Point
from openquake.hazardlib.geo.utils import OrthographicProjection


# local imports
from esi_utils_rupture.base import Rupture
from esi_utils_rupture.utils import (
    _quad_distance,
    get_quad_width,
    get_quad_normal,
    get_vertical_vector,
    get_quad_length,
    is_quad,
)
from esi_utils_rupture import gc2
import esi_utils_rupture.constants as constants
from esi_utils_time.ancient_time import HistoricTime
from esi_utils_vectors.ecef import latlon2ecef
from esi_utils_vectors.ecef import ecef2latlon
from esi_utils_vectors.vector import Vector


class QuadRupture(Rupture):
    """
    Rupture class that represents the rupture surface as a combination of
    quadrilaterals. Each quadrilateral must have horizontal top and bottom
    edges and must be coplanar. These restrictions make the computation of
    rupture distances more efficient. The number of points in the top edges
    must match the number of points in the bottom edge.
    """

    def __init__(self, d, origin):
        """
        Create a QuadRupture instance from a GeoJSON dictionary and an Origin.

        Args:
           d (dict): Rupture GeoJSON dictionary.
           origin (Origin): Reference to a ShakeMap Origin object.

        Returns:
            QuadRupture instance.

        """

        polys = d["features"][0]["geometry"]["coordinates"][0]
        n_polygons = len(polys)
        lon = []
        lat = []
        dep = []
        for i in range(n_polygons):
            p = polys[i]
            p_lons = [pt[0] for pt in p][0:-1]
            p_lats = [pt[1] for pt in p][0:-1]
            p_depths = [pt[2] for pt in p][0:-1]
            lon = lon + p_lons + [np.nan]
            lat = lat + p_lats + [np.nan]
            dep = dep + p_depths + [np.nan]

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, datetime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v

        self._geojson = d
        self._lon = lon
        self._lat = lat
        self._depth = dep
        self._origin = origin
        self._reference = d["metadata"]["reference"]
        self._setQuadrilaterals()

    def getDepthAtPoint(self, lat, lon):
        SMALL_DISTANCE = 2e-03  # 2 meters
        depth = np.nan

        tmp = self.computeRjb(np.array([lon]), np.array([lat]), np.array([0]))
        if tmp > SMALL_DISTANCE:
            return depth

        i = 0
        imin = -1
        dmin = 9999999999999999
        for quad in self.getQuadrilaterals():
            pX = Vector.fromPoint(Point(lon, lat, 0))
            points = np.reshape(np.array([pX.x, pX.y, pX.z]), (1, 3))
            rjb = _quad_distance(quad, points, horizontal=True)
            if rjb[0][0] < dmin:
                dmin = rjb[0][0]
                imin = i
            i += 1

        quad = self._quadrilaterals[imin]
        P0, P1, P2, P3 = quad
        # project the quad and the point in question to orthographic defined by
        # quad
        xmin = np.min([P0.x, P1.x, P2.x, P3.x])
        xmax = np.max([P0.x, P1.x, P2.x, P3.x])
        ymin = np.min([P0.y, P1.y, P2.y, P3.y])
        ymax = np.max([P0.y, P1.y, P2.y, P3.y])
        proj = OrthographicProjection(xmin, xmax, ymax, ymin)

        # project each vertex of quad (at 0 depth)
        s0x, s0y = proj(P0.x, P0.y)
        s1x, s1y = proj(P1.x, P1.y)
        s2x, s2y = proj(P2.x, P2.y)
        s3x, s3y = proj(P3.x, P3.y)
        sxx, sxy = proj(lon, lat)

        # turn these to vectors
        s0 = Vector(s0x, s0y, 0)
        s1 = Vector(s1x, s1y, 0)
        s3 = Vector(s3x, s3y, 0)
        sx = Vector(sxx, sxy, 0)

        # Compute vector from s0 to s1
        s0s1 = s1 - s0
        # Compute the vector from s0 to s3
        s0s3 = s3 - s0
        # Compute the vector from s0 to sx
        s0sx = sx - s0

        # cross products
        s0normal = s0s3.cross(s0s1)
        dd = s0s1.cross(s0normal)
        # normalize dd (down dip direction)
        ddn = dd.norm()
        # dot product
        sxdd = ddn.dot(s0sx)

        # get width of quad (convert from km to m)
        w = get_quad_width(quad) * 1000

        # Get weights for top and bottom edge depths
        N = get_quad_normal(quad)
        V = get_vertical_vector(quad)
        dip = np.degrees(np.arccos(Vector.dot(N, V)))
        ws = w * np.cos(np.radians(dip))
        wtt = (ws - sxdd) / ws
        wtb = sxdd / ws

        # Compute the depth of of the plane at Px:
        depth = wtt * P0.z + wtb * P3.z * 1000

        return depth

    def getLength(self):
        """
        Compute length of rupture based on top edge in km.

        Returns:
            float: Length of rupture (km).

        """
        flength = 0
        for quad in self._quadrilaterals:
            flength = flength + get_quad_length(quad)
        return flength

    def getWidth(self):
        """
        Compute average rupture width (km) for all quadrilaterals defined for
        the rupture.

        Returns:
            float: Average width in km of all rupture quadrilaterals.
        """
        wsum = 0.0
        for quad in self._quadrilaterals:
            wsum = wsum + get_quad_width(quad)
        mwidth = wsum / len(self._quadrilaterals)
        return mwidth

    def getArea(self):
        """
        Compute area of rupture.

        Returns:
            float: Rupture area in square km.

        """
        asum = 0.0
        for quad in self._quadrilaterals:
            width = get_quad_width(quad)
            length = get_quad_length(quad)
            asum = asum + width * length
        return asum

    @classmethod
    def fromTrace(
        cls,
        xp0,
        yp0,
        xp1,
        yp1,
        zp,
        widths,
        dips,
        origin,
        strike=None,
        group_index=None,
        reference="",
    ):
        """
        Create a QuadRupture instance from a set of vertices that define the
        top of the rupture, and an array of widths/dips.

        Each rupture quadrilaterial is defined by specifying the latitude,
        longitude, and depth of the two vertices on the top edges, which must
        have the dame depths. The other verticies are then constructed from
        the top edges and the width and dip of the quadrilateral.

        Args:
            xp0 (array): Array or list of longitudes (floats) of p0.
            yp0 (array): Array or list of latitudes (floats) of p0.
            xp1 (array): Array or list of longitudes (floats) of p1.
            yp1 (array): Array or list of latitudes (floats) of p1.
            zp (array): Array or list of depths for each of the top of rupture
                rectangles (km).
            widths (array): Array of widths for each of rectangle (km).
            dips (array): Array of dips for each of rectangle (degrees).
            origin (Origin): Reference to a ShakeMap origin object.
            strike (array): If None then strike is computed from verticies of
                top edge of each quadrilateral. If the array has only a
                single value, then all
                quadrilaterals are constructed assuming this strike direction.
                If an array with the same length as the trace coordinates then
                it specifies the strike for each quadrilateral.
            group_index (list): List of integers to indicate group index. If
                None then each quadrilateral is assumed to be in a different
                group since there is no guarantee that any of them are
                continuous.
            reference (str): String explaining where the rupture definition
                came from (publication style reference, etc.).

        Returns:
            QuadRupture instance.

        Raises:
            Exception: if the input arrays are not all the same
                length, or if the strike array is not length 1 or the same
                length as the input arrays.

        """
        if not (
            len(xp0)
            == len(yp0)
            == len(xp1)
            == len(yp1)
            == len(zp)
            == len(dips)
            == len(widths)
        ):
            raise ValueError(
                "Number of xp0, yp0, xp1, yp1, zp, widths, dips points must be equal."
            )
        if strike is not None and len(xp0) != len(strike) and len(strike) != 1:
            raise ValueError(
                "Strike must be None or an array of one value or the "
                "same length as trace coordinates."
            )

        if group_index is None:
            group_index = np.array(range(len(xp0)))

        # Convert dips to radians
        dips = np.radians(dips)

        # Ensure that all input sequences are numpy arrays
        xp0 = np.array(xp0, dtype="d")
        xp1 = np.array(xp1, dtype="d")
        yp0 = np.array(yp0, dtype="d")
        yp1 = np.array(yp1, dtype="d")
        zp = np.array(zp, dtype="d")
        widths = np.array(widths, dtype="d")
        dips = np.array(dips, dtype="d")

        # Get a projection object
        west = np.min((xp0.min(), xp1.min()))
        east = np.max((xp0.max(), xp1.max()))
        south = np.min((yp0.min(), yp1.min()))
        north = np.max((yp0.max(), yp1.max()))

        # Projected coordinates are in km
        proj = OrthographicProjection(west, east, north, south)
        xp2 = np.zeros_like(xp0)
        xp3 = np.zeros_like(xp0)
        yp2 = np.zeros_like(xp0)
        yp3 = np.zeros_like(xp0)
        zpdown = np.zeros_like(zp)
        for i in range(0, len(xp0)):
            # Project the top edge coordinates
            p0x, p0y = proj(xp0[i], yp0[i])
            p1x, p1y = proj(xp1[i], yp1[i])

            # Get the rotation angle defined by these two points
            if strike is None:
                dx = p1x - p0x
                dy = p1y - p0y
                theta = np.arctan2(dx, dy)  # theta is angle from north
            elif len(strike) == 1:
                theta = np.radians(strike[0])
            else:
                theta = np.radians(strike[i])

            R = np.array(
                [
                    [np.cos(theta), -np.sin(theta)],
                    [np.sin(theta), np.cos(theta)],
                ]
            )

            # Rotate the top edge points into a new coordinate system (vertical
            # line)
            p0 = np.array([p0x, p0y])
            p1 = np.array([p1x, p1y])
            p0p = np.dot(R, p0)
            p1p = np.dot(R, p1)

            # Get right side coordinates in project, rotated system
            dz = np.sin(dips[i]) * widths[i]
            dx = np.cos(dips[i]) * widths[i]
            p3xp = p0p[0] + dx
            p3yp = p0p[1]
            p2xp = p1p[0] + dx
            p2yp = p1p[1]

            # Get right side coordinates in un-rotated projected system
            p3p = np.array([p3xp, p3yp])
            p2p = np.array([p2xp, p2yp])
            Rback = np.array(
                [
                    [np.cos(-theta), -np.sin(-theta)],
                    [np.sin(-theta), np.cos(-theta)],
                ]
            )
            p3 = np.dot(Rback, p3p)
            p2 = np.dot(Rback, p2p)
            p3x = np.array([p3[0]])
            p3y = np.array([p3[1]])
            p2x = np.array([p2[0]])
            p2y = np.array([p2[1]])

            # project lower edge points back to lat/lon coordinates
            lon3, lat3 = proj(p3x, p3y, reverse=True)
            lon2, lat2 = proj(p2x, p2y, reverse=True)

            xp2[i] = lon2[0]
            xp3[i] = lon3[0]
            yp2[i] = lat2[0]
            yp3[i] = lat3[0]
            zpdown[i] = zp[i] + dz

        # ---------------------------------------------------------------------
        # Create GeoJSON object
        # ---------------------------------------------------------------------

        coords = []
        u_groups = np.unique(group_index)
        n_groups = len(u_groups)
        for i in range(n_groups):
            ind = np.where(u_groups[i] == group_index)[0]
            lons = np.concatenate(
                [
                    xp0[ind[0]].reshape((1,)),
                    xp1[ind],
                    xp2[ind][::-1],
                    xp3[ind][::-1][-1].reshape((1,)),
                    xp0[ind[0]].reshape((1,)),
                ]
            )
            lats = np.concatenate(
                [
                    yp0[ind[0]].reshape((1,)),
                    yp1[ind],
                    yp2[ind][::-1],
                    yp3[ind][::-1][-1].reshape((1,)),
                    yp0[ind[0]].reshape((1,)),
                ]
            )
            deps = np.concatenate(
                [
                    zp[ind[0]].reshape((1,)),
                    zp[ind],
                    zpdown[ind][::-1],
                    zpdown[ind][::-1][-1].reshape((1,)),
                    zp[ind[0]].reshape((1,)),
                ]
            )

            poly = []
            for lon, lat, dep in zip(lons, lats, deps):
                poly.append([lon, lat, dep])
            coords.append(poly)

        d = {
            "type": "FeatureCollection",
            "metadata": {"reference": reference},
            "features": [
                {
                    "type": "Feature",
                    "properties": {"rupture type": "rupture extent"},
                    "geometry": {
                        "type": "MultiPolygon",
                        "coordinates": [coords],
                    },
                }
            ],
        }

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, HistoricTime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v

        return cls(d, origin)

    def writeTextFile(self, rupturefile):
        """
        Write rupture data to rupture file format as defined in ShakeMap
        Software Guide.

        Note that this currently treats each quadrilateral as a separate
        polygon. This needs to be udpated.

        Args:
            rupturefile (str): Filename of output data file OR file-like
                object.

        """
        if not hasattr(rupturefile, "read"):
            f = open(rupturefile, "wt")
        else:
            f = rupturefile  # just a reference to the input file-like object
        f.write(f"#{self._reference}\n")
        for quad in self.getQuadrilaterals():
            P0, P1, P2, P3 = quad
            f.write(f"{P0.latitude:.4f} {P0.longitude:.4f} {P0.depth:.4f}\n")
            f.write(f"{P1.latitude:.4f} {P1.longitude:.4f} {P1.depth:.4f}\n")
            f.write(f"{P2.latitude:.4f} {P2.longitude:.4f} {P2.depth:.4f}\n")
            f.write(f"{P3.latitude:.4f} {P3.longitude:.4f} {P3.depth:.4f}\n")
            f.write(f"{P0.latitude:.4f} {P0.longitude:.4f} {P0.depth:.4f}\n")
            f.write(">\n")
        if not hasattr(rupturefile, "read"):
            f.close()

    @classmethod
    def fromOrientation(
        cls, px, py, pz, dx, dy, length, width, strike, dip, origin
    ):
        """
        Create a QuadRupture instance from a known point, shape, and
        orientation.

        A point is defined as a set of latitude, longitude, and depth, which
        is located in the corner between the tail of the vector pointing in
        the strike direction and the dip direction (nearest to the surface).
        The shape is defined by length, width, dx, and dy. The length is the
        measurement of the quadrilateral in the direction of strike, and
        width is the measurement of quadrilateral in the direction of dip.
        Dx is the measurement on the plane in the strike direction between
        the known point and the corner between the tail of the vector pointing
        in the strike direction and the dip direction (nearest to the surface).
        Dy is the measurement on the plane in the dip direction between
        the known point and the corner between the tail of the vector pointing
        in the strike direction and the dip direction (nearest to the surface).
        The orientation is defined by azimuth and angle from
        horizontal, strike and dip respectively. For example in plane view:
            ::
                            strike direction
                        p1*------------------->>p2
                        *        | dy           |
                 dip    |--------o              |
              direction |   dx    known point   | Width
                        V                       |
                        V                       |
                        p4----------------------p3
                                Length

        Args:
            px (array): Array or list of longitudes (floats) of the known
                point.
            py (array): Array or list of latitudes (floats) of the known point.
            pz (array): Array or list of depths (floats) of the known point.
            dx (array): Array or list of distances (floats), in the strike
                direction, between the known point and P1. dx must be less than
                length.
            dy (array): Array or list of distances (floats), in the dip
                direction, between the known point and P1. dy must be less than
                width.
            length (array): Array or list of widths (floats) of the plane in
                the strike direction.
            width (array): Array or list of widths (floats) of the plane in the
                dip direction.
            strike (array): Array or list of strike angles (floats).
            dip (array): Array or list of dip angles (floats).
            origin (Origin): Reference to a ShakeMap Origin object.
        Returns:
            QuadRupture instance.

        Raises:
            Exception: if the lengths of the points arrays are not
                all equal.
        """
        # Verify that arrays are of equal length
        if (
            len(px)
            == len(py)
            == len(pz)
            == len(dx)
            == len(dy)
            == len(length)
            == len(width)
            == len(strike)
            == len(dip)
        ):
            pass
        else:
            raise ValueError(
                "Number of px, py, pz, dx, dy, length, width, strike, dip points must "
                "be equal."
            )

        # Verify that all are numpy arrays
        px = np.array(px, dtype="d")
        py = np.array(py, dtype="d")
        pz = np.array(pz, dtype="d")
        dx = np.array(dx, dtype="d")
        dy = np.array(dy, dtype="d")
        length = np.array(length, dtype="d")
        width = np.array(width, dtype="d")
        strike = np.array(strike, dtype="d")
        dip = np.array(dip, dtype="d")

        # Get P1 and P2 (top horizontal points)
        theta = np.rad2deg(np.arctan2(dy * np.cos(np.deg2rad(dip)), dx))
        P1_direction = strike + 180 + theta
        P1_distance = np.sqrt(dx**2 + (dy * np.cos(np.deg2rad(dip))) ** 2)
        P2_direction = strike
        P2_distance = length
        P1_lon = []
        P1_lat = []
        P2_lon = []
        P2_lat = []
        for idx, value in enumerate(px):
            P1_points = point_at(
                px[idx], py[idx], P1_direction[idx], P1_distance[idx]
            )
            P1_lon += [P1_points[0]]
            P1_lat += [P1_points[1]]
            P2_points = point_at(
                P1_points[0], P1_points[1], P2_direction[idx], P2_distance[idx]
            )
            P2_lon += [P2_points[0]]
            P2_lat += [P2_points[1]]

        # Get top depth
        top_horizontal_depth = pz - np.abs(dy * np.sin(np.deg2rad(dip)))

        # Get QuadRupture object
        quad = QuadRupture.fromTrace(
            P1_lon,
            P1_lat,
            P2_lon,
            P2_lat,
            top_horizontal_depth,
            width,
            dip,
            origin,
            strike=strike,
        )
        return quad

    @classmethod
    def fromVertices(
        cls,
        xp0,
        yp0,
        zp0,
        xp1,
        yp1,
        zp1,
        xp2,
        yp2,
        zp2,
        xp3,
        yp3,
        zp3,
        origin,
        group_index=None,
        reference=None,
    ):
        """
        Create a QuadDrupture instance from the vector of vertices that fully
        define the quadrilaterals. The points p0, ..., p3 are labeled below for
        a trapezoid:

        ::

              p0--------p1
             /          |
            /           |
           p3-----------p2

        All of the following vector arguments must have the same length.

        Args:
            xp0 (array): Array or list of longitudes (floats) of p0.
            yp0 (array): Array or list of latitudes (floats) of p0.
            zp0 (array): Array or list of depths (floats) of p0.
            xp1 (array): Array or list of longitudes (floats) of p1.
            yp1 (array): Array or list of latitudes (floats) of p1.
            zp1 (array): Array or list of depths (floats) of p1.
            xp2 (array): Array or list of longitudes (floats) of p2.
            yp2 (array): Array or list of latitudes (floats) of p2.
            zp2 (array): Array or list of depths (floats) of p2.
            xp3 (array): Array or list of longitudes (floats) of p3.
            yp3 (array): Array or list of latitudes (floats) of p3.
            zp3 (array): Array or list of depths (floats) of p3.
            origin (Origin): Reference to a ShakeMap Origin object.
            group_index (list): List of integers to indicate group index. If
                None then each quadrilateral is assumed to be in a different
                group since there is no guarantee that any of them are
                continuous.
            reference (str): String explaining where the rupture definition
                came from (publication style reference, etc.)

        Returns:
            QuadRupture object, where the rupture is modeled as a series of
            trapezoids.

        Raises:
            Exception: if the lengths of the inptu arrays are not
                all equal, or if the length of the group_index is not the
                same as the arrays (if group_index is supplied)..
        """
        if (
            len(xp0)
            == len(yp0)
            == len(zp0)
            == len(xp1)
            == len(yp1)
            == len(zp1)
            == len(xp2)
            == len(yp2)
            == len(zp2)
            == len(xp3)
            == len(yp3)
            == len(zp3)
        ):
            pass
        else:
            raise ValueError(
                "All vectors specifying quadrilateral vertices must have the same "
                "length."
            )

        nq = len(xp0)
        if group_index is not None:
            if len(group_index) != nq:
                raise ValueError(
                    "group_index must have same length as vertices."
                )
        else:
            group_index = np.array(range(nq))

        xp0 = np.array(xp0, dtype="d")
        yp0 = np.array(yp0, dtype="d")
        zp0 = np.array(zp0, dtype="d")
        xp1 = np.array(xp1, dtype="d")
        yp1 = np.array(yp1, dtype="d")
        zp1 = np.array(zp1, dtype="d")
        xp2 = np.array(xp2, dtype="d")
        yp2 = np.array(yp2, dtype="d")
        zp2 = np.array(zp2, dtype="d")
        xp3 = np.array(xp3, dtype="d")
        yp3 = np.array(yp3, dtype="d")
        zp3 = np.array(zp3, dtype="d")

        # ---------------------------------------------------------------------
        # Create GeoJSON object
        # ---------------------------------------------------------------------

        coords = []
        u_groups = np.unique(group_index)
        n_groups = len(u_groups)
        for i in range(n_groups):
            ind = np.where(u_groups[i] == group_index)[0]
            lons = np.concatenate(
                [
                    xp0[ind[0]].reshape((1,)),
                    xp1[ind],
                    xp2[ind][::-1],
                    xp3[ind][::-1][-1].reshape((1,)),
                    xp0[ind[0]].reshape((1,)),
                ]
            )
            lats = np.concatenate(
                [
                    yp0[ind[0]].reshape((1,)),
                    yp1[ind],
                    yp2[ind][::-1],
                    yp3[ind][::-1][-1].reshape((1,)),
                    yp0[ind[0]].reshape((1,)),
                ]
            )
            deps = np.concatenate(
                [
                    zp0[ind[0]].reshape((1,)),
                    zp1[ind],
                    zp2[ind][::-1],
                    zp3[ind][::-1][-1].reshape((1,)),
                    zp0[ind[0]].reshape((1,)),
                ]
            )

            poly = []
            for lon, lat, dep in zip(lons, lats, deps):
                poly.append([lon, lat, dep])
            coords.append(poly)

        d = {
            "type": "FeatureCollection",
            "metadata": {"reference": reference},
            "features": [
                {
                    "type": "Feature",
                    "properties": {"rupture type": "rupture extent"},
                    "geometry": {
                        "type": "MultiPolygon",
                        "coordinates": [coords],
                    },
                }
            ],
        }

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, HistoricTime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v
        if hasattr(origin, "id"):
            d["metadata"]["eventid"] = origin.id

        return cls(d, origin)

    def getQuadrilaterals(self):
        """
        Return a list of quadrilaterals.

        Returns:
            list: List of quadrilaterals where each quad is a tuple of four
                `Point <https://github.com/gem/oq-hazardlib/blob/master/openquake/hazardlib/geo/point.py>`__
                objects.
        """  # noqa
        return copy.deepcopy(self._quadrilaterals)

    def getStrike(self):
        """
        Return strike angle. If rupture consists of multiple quadrilaterals,
        the average strike angle, weighted by quad length, is returned.
        Note: for ruptures with quads where the strike angle changes by 180 deg
        due to reverses in dip direction are problematic and not handeled well
        by this algorithm.

        Returns:
            float: Strike angle in degrees.

        """
        nq = len(self._quadrilaterals)
        strikes = np.zeros(nq)
        lengths = np.zeros(nq)
        for i in range(nq):
            P0 = self._quadrilaterals[i][0]
            P1 = self._quadrilaterals[i][1]
            strikes[i] = P0.azimuth(P1)
            lengths[i] = get_quad_length(self._quadrilaterals[i])
        x = np.sin(np.radians(strikes))
        y = np.cos(np.radians(strikes))
        xbar = np.sum(x * lengths) / np.sum(lengths)
        ybar = np.sum(y * lengths) / np.sum(lengths)
        return np.degrees(np.arctan2(xbar, ybar))

    def getDepthToTop(self):
        """
        Determine shallowest vertex of entire rupture.

        :returns:
            Shallowest depth of all vertices (float).
        """
        mindep = 9999999
        for quad in self._quadrilaterals:
            P0, P1, P2, P3 = quad
            depths = np.array([P0.depth, P1.depth, P2.depth, P3.depth])
            if np.min(depths) < mindep:
                mindep = np.min(depths)
        return mindep

    def getDip(self):
        """
        Return average dip of all quadrilaterals in the rupture.

        Returns:
           float: Average dip in degrees.

        """
        dipsum = 0.0
        for quad in self._quadrilaterals:
            N = get_quad_normal(quad)
            V = get_vertical_vector(quad)
            dipsum = dipsum + np.degrees(np.arccos(Vector.dot(N, V)))
        dip = dipsum / len(self._quadrilaterals)
        return dip

    def getIndividualWidths(self):
        """
        Return an array of rupture widths (km), one for each quadrilateral
        defined for the rupture.

        Returns:
            Array of quad widths in km of all rupture quadrilaterals.
        """
        nquad = self.getNumQuads()
        widths = np.zeros(nquad)
        for i in range(nquad):
            q = self._quadrilaterals[i]
            widths[i] = get_quad_width(q)
        return widths

    def getIndividualTopLengths(self):
        """
        Return an array of rupture lengths along top edge (km),
        one for each quadrilateral defined for the rupture.

        Returns:
            Array of lengths in km of top edge of quadrilaterals.
        """
        nquad = self.getNumQuads()
        lengths = np.zeros(nquad)
        for i in range(nquad):
            P0, P1, P2, P3 = self._quadrilaterals[i]
            p0 = Vector.fromPoint(P0)
            p1 = Vector.fromPoint(P1)
            lengths[i] = (p1 - p0).mag() / 1000.0
        return lengths

    @staticmethod
    def _fixStrikeDirection(quad):
        P0, P1, P2, P3 = quad
        eps = 1e-6
        p0 = Vector.fromPoint(P0)  # fromPoint converts to ECEF
        p1 = Vector.fromPoint(P1)
        p2 = Vector.fromPoint(P2)
        p1p0 = p1 - p0
        p2p0 = p2 - p0
        qnv = Vector.cross(p2p0, p1p0).norm()
        tmp = p0 + qnv
        tmplat, tmplon, tmpz = ecef2latlon(tmp.x, tmp.y, tmp.z)
        if (tmpz - P0.depth) < eps:  # If True then do nothing
            fixed = quad
        else:
            newP0 = copy.deepcopy(P1)
            newP1 = copy.deepcopy(P0)
            newP2 = copy.deepcopy(P3)
            newP3 = copy.deepcopy(P2)
            fixed = [newP0, newP1, newP2, newP3]
        return fixed

    def _setQuadrilaterals(self):
        """
        Create internal list of N quadrilaterals. Reverses quad if dip
        direction is incorrect.
        """

        # Make sure arrays are numpy arrays.
        self._lon = np.array(self._lon)
        self._lat = np.array(self._lat)
        self._depth = np.array(self._depth)

        # Find the nans, which tells is where the separate polygons/groups are
        group_ends = np.where(np.isnan(self._lon))[0]
        n_groups = len(group_ends)

        # Check that arrays are the same length
        if len(self._lon) != len(self._lat) != len(self._depth):
            raise IndexError(
                "Length of input lon, lat, depth arrays must be equal"
            )

        # Construct quads
        group_start = 0

        self._quadrilaterals = []
        self._group_index = []
        groupind = 0
        for i in range(n_groups):
            lonseg = self._lon[group_start : group_ends[i]]
            latseg = self._lat[group_start : group_ends[i]]
            depthseg = self._depth[group_start : group_ends[i]]

            # Each group can have many contiguous quadrilaterals defined in it
            # separations (nans) between segments mean that segments are not
            # contiguous.

            npoints = len(lonseg)
            nquads = int((npoints - 4) / 2) + 1
            quad_start = 0
            quad_end = -1
            for j in range(nquads):
                P0 = Point(
                    lonseg[quad_start],
                    latseg[quad_start],
                    depthseg[quad_start],
                )
                P1 = Point(
                    lonseg[quad_start + 1],
                    latseg[quad_start + 1],
                    depthseg[quad_start + 1],
                )
                P2 = Point(
                    lonseg[quad_end - 1],
                    latseg[quad_end - 1],
                    depthseg[quad_end - 1],
                )
                P3 = Point(
                    lonseg[quad_end], latseg[quad_end], depthseg[quad_end]
                )
                quad = [P0, P1, P2, P3]

                # Enforce plane by moving P2 -- already close because of check
                # in read_rupture_file/is_quadrupture_class/is_quad

                dummy, fixed_quad = is_quad(quad)

                # Reverse quad if necessary
                fixed_quad = self._fixStrikeDirection(fixed_quad)

                self._quadrilaterals.append(fixed_quad)

                quad_start = quad_start + 1
                quad_end = quad_end - 1

            group_start = group_ends[i] + 1
            self._group_index.extend([groupind] * nquads)
            groupind = groupind + 1

    def _getGroupIndex(self):
        """
        Return a list of segment group indexes.

        Returns:
            list: Segment group indexes; length equals the number of
                quadrilaterals.
        """
        return copy.deepcopy(self._group_index)

    @property
    def lats(self):
        """
        Return an array of latitudes for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop latitude values; disconnected
                segments are separated by nans.
        """
        lats = []
        quads = self.getQuadrilaterals()
        groups = self._getGroupIndex()
        u_groups = np.unique(groups)
        ng = len(u_groups)
        for i in range(ng):
            q_ind = np.where(groups == u_groups[i])[0]
            nq = len(q_ind)
            top_lats = []
            bot_lats = []
            for j in range(nq):
                if j == 0:
                    top0 = [quads[q_ind[j]][0].latitude]
                    bot0 = [quads[q_ind[j]][3].latitude]
                    top_lats = top_lats + top0
                    bot_lats = bot_lats + bot0
                top_lats = top_lats + [quads[q_ind[j]][1].latitude]
                bot_lats = bot_lats + [quads[q_ind[j]][2].latitude]
            lats = lats + top_lats + bot_lats[::-1] + top0 + [np.nan]

        return np.array(lats)

    @property
    def lons(self):
        """
        Return an array of longitudes for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop longitude values; disconnected
                segments are separated by nans.
        """
        lons = []
        quads = self.getQuadrilaterals()
        groups = self._getGroupIndex()
        u_groups = np.unique(groups)
        ng = len(u_groups)
        for i in range(ng):
            q_ind = np.where(groups == u_groups[i])[0]
            nq = len(q_ind)
            top_lons = []
            bot_lons = []
            for j in range(nq):
                if j == 0:
                    top0 = [quads[q_ind[j]][0].longitude]
                    bot0 = [quads[q_ind[j]][3].longitude]
                    top_lons = top_lons + top0
                    bot_lons = bot_lons + bot0
                top_lons = top_lons + [quads[q_ind[j]][1].longitude]
                bot_lons = bot_lons + [quads[q_ind[j]][2].longitude]
            lons = lons + top_lons + bot_lons[::-1] + top0 + [np.nan]
        return np.array(lons)

    @property
    def depths(self):
        """
        Return an array of depths for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop depths; disconnected
                segments are separated by nans.
        """
        deps = []
        quads = self.getQuadrilaterals()
        groups = self._getGroupIndex()
        u_groups = np.unique(groups)
        ng = len(u_groups)
        for i in range(ng):
            q_ind = np.where(groups == u_groups[i])[0]
            nq = len(q_ind)
            top_deps = []
            bot_deps = []
            for j in range(nq):
                if j == 0:
                    top0 = [quads[q_ind[j]][0].depth]
                    bot0 = [quads[q_ind[j]][3].depth]
                    top_deps = top_deps + top0
                    bot_deps = bot_deps + bot0
                top_deps = top_deps + [quads[q_ind[j]][1].depth]
                bot_deps = bot_deps + [quads[q_ind[j]][2].depth]
            deps = deps + top_deps + bot_deps[::-1] + top0 + [np.nan]

        return np.array(deps)

    def getDeps(self):
        """
        Return a copy of the array of depths for the rupture verticies.

        Returns:
            array: Numpy array of latitude values.
        """
        return self._depth.copy()

    def getNumGroups(self):
        """
        Return a count of the number of rupture groups.

        Returns:
            int:Rnumber of rupture groups.

        """
        return len(np.unique(self._group_index))

    def getNumQuads(self):
        """
        Return a count of the number of rupture quadrilaterals.

        Returns:
            int: Number of rupture quadrilaterals.
        """
        return len(self._quadrilaterals)

    def getRuptureAsArrays(self):
        """
        Return a 3-tuple of numpy arrays indicating X, Y, Z (lon,lat,depth)
        coordinates. Rupture groups are separated by numpy.NaN values.

        Returns:
            tuple: 3-tuple of numpy arrays indicating X,Y,Z (lon,lat,depth)
                coordinates.
        """
        return (
            np.array(self._lon),
            np.array(self._lat),
            np.array(self._depth),
        )

    def getRuptureAsMesh(self):
        """
        Return rupture segments as a OQ-Hazardlib Mesh object.

        Returns:
            Mesh (https://github.com/gem/oq-hazardlib/blob/master/openquake/hazardlib/geo/mesh.py)
        """  # noqa
        rupture = Mesh(self._lon, self._lat, self._depth)
        return rupture

    def computeRjb(self, lon, lat, depth):
        """
        Method for computing Joyner-Boore distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           tuple: A tuple of an array of Joyner-Boore distance (km), and None.

        """

        # ---------------------------------------------------------------------
        # Sort out sites
        # ---------------------------------------------------------------------
        oldshape = lon.shape

        if len(oldshape) == 2:
            newshape = (oldshape[0] * oldshape[1], 1)
        else:
            newshape = (oldshape[0], 1)

        x, y, z = latlon2ecef(lat, lon, depth)
        x.shape = newshape
        y.shape = newshape
        z.shape = newshape
        sites_ecef = np.hstack((x, y, z))

        minrjb = np.ones(newshape, dtype=lon.dtype) * 1e16
        quads = self.getQuadrilaterals()

        for q in quads:
            p0, p1, p2, p3 = q
            s0 = copy.deepcopy(p0)
            s1 = copy.deepcopy(p1)
            s2 = copy.deepcopy(p2)
            s3 = copy.deepcopy(p3)
            s0.depth = 0.0
            s1.depth = 0.0
            s2.depth = 0.0
            s3.depth = 0.0
            squad = [s0, s1, s2, s3]
            rjbdist = _quad_distance(squad, sites_ecef, horizontal=True)
            minrjb = np.minimum(minrjb, rjbdist)

        minrjb = minrjb.reshape(oldshape)
        return minrjb

    def computeRrup(self, lon, lat, depth):
        """
        Method for computing rupture distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           tuple: A tuple of an array of Rupture distance (km), and None.

        """

        # ---------------------------------------------------------------------
        # Sort out sites
        # ---------------------------------------------------------------------
        oldshape = lon.shape

        if len(oldshape) == 2:
            newshape = (oldshape[0] * oldshape[1], 1)
        else:
            newshape = (oldshape[0], 1)

        x, y, z = latlon2ecef(lat, lon, depth)
        x.shape = newshape
        y.shape = newshape
        z.shape = newshape
        sites_ecef = np.hstack((x, y, z))

        minrrup = np.ones(newshape, dtype=lon.dtype) * 1e16
        quads = self.getQuadrilaterals()

        for q in quads:
            rrupdist = _quad_distance(q, sites_ecef)
            minrrup = np.minimum(minrrup, rrupdist)

        minrrup = minrrup.reshape(oldshape)
        return minrrup

    def computeGC2(self, lon, lat, depth):
        """
        Method for computing version 2 of the Generalized Coordinate system
        (GC2) by Spudich and Chiou OFR 2015-1028.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            dict: Dictionary with keys for each of the GC2-related distances,
                which include 'rx', 'ry', 'ry0', 'U', 'T'.
        """
        # This just hands off to the module-level method
        dict = gc2._computeGC2(self, lon, lat, depth)
        return dict



#!/usr/bin/env python

# stdlib imports
import os
import os.path
import io
import copy
import tempfile
import shutil
import time
import datetime

# third party
import numpy as np
import pytest
from obspy.core.event import Catalog, FocalMechanism, Event
from obspy.core.event.source import (
    NodalPlane,
    NodalPlanes,
    Axis,
    PrincipalAxes,
)

from esi_utils_rupture.origin import Origin, read_moment_quakeml
from esi_utils_rupture.quad_rupture import QuadRupture
from esi_utils_rupture.edge_rupture import EdgeRupture
from esi_utils_rupture.factory import (
    get_rupture,
    text_to_json,
    rupture_from_dict,
    validate_json,
)

from esi_utils_rupture.utils import (
    get_local_unit_slip_vector,
    get_local_unit_slip_vector_DS,
    get_local_unit_slip_vector_SS,
)
from esi_utils_rupture.utils import (
    get_quad_slip,
    get_quad_strike_vector,
    get_quad_down_dip_vector,
    rake_to_mech,
)
from esi_utils_rupture import constants
from esi_utils_time.ancient_time import HistoricTime

homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?

do_tests = True


def test_no_origin():
    # No argument
    with pytest.raises(Exception):
        get_rupture()

    # Wrong type
    with pytest.raises(Exception):
        get_rupture(7.3)


def test_text_to_json():
    sm_data = """#Oglesby, D. D., D. S. Dreger, R. A. Harris, \
N. Ratchkovski, and R. Hansen (2004). Inverse kinematic and forward dynamic \
models of the 2002 Denali fault earthquake, Alaska, Bull. Seism. Soc. Am. 94, \
S214-S233.
>
-147.807 63.434        0.000
-147.210 63.472        0.000
-147.267 63.650        22.294
-147.864 63.613        22.294
-147.807 63.434        0.000
>
-146.951 63.551        0.000
-147.551 63.518        0.000
-147.551 63.518        30.000
-146.951 63.551        30.000
-146.951 63.551        0.000
>
-145.968 63.453        0.000
-146.952 63.547        0.000
-146.952 63.547        30.000
-145.968 63.453        30.000
-145.968 63.453        0.000
>
-143.586 62.872        0.000
-145.996 63.427        0.000
-145.996 63.427        30.000
-143.586 62.872        30.000
-143.586 62.872        0.000
>
-142.500 62.114        0.000
-143.669 62.831        0.000
-143.669 62.831        30.000
-142.500 62.114        30.000
-142.500 62.114        0.000"""
    stringio = io.StringIO(sm_data)
    jdict = text_to_json(stringio)
    refcmp = """Oglesby, D. D., D. S. Dreger, R. A. Harris, \
N. Ratchkovski, and R. Hansen (2004). Inverse kinematic and forward dynamic \
models of the 2002 Denali fault earthquake, Alaska, Bull. Seism. Soc. Am. 94, \
S214-S233."""
    #    assert jdict['metadata']['reference'] == refcmp.replace('\n', ' ')
    assert jdict["metadata"]["reference"] == refcmp
    cstart = [
        [-147.807, 63.434, 0.0],
        [-147.21, 63.472, 0.0],
        [-147.267, 63.65, 22.294],
        [-147.864, 63.613, 22.294],
        [-147.807, 63.434, 0.0],
    ]
    cend = [
        [-142.5, 62.114, 0.0],
        [-143.669, 62.831, 0.0],
        [-143.669, 62.831, 30.0],
        [-142.5, 62.114, 30.0],
        [-142.5, 62.114, 0.0],
    ]

    assert len(jdict["features"][0]["geometry"]["coordinates"][0]) == 5
    assert jdict["features"][0]["geometry"]["coordinates"][0][0] == cstart
    assert jdict["features"][0]["geometry"]["coordinates"][0][-1] == cend

    sm_data2 = """#Source: NEIC (2004) based on aftershock distribution
2.98 94.45 10.0
5.0  94.0 10.0
6.0  93.5 10.0
7.0  93.2 10.0
8.0  93.1 10.0
10.0 92.5 10.0
12.0 92.5 10.0
12.0 93.5 40.0
10.0 93.5 40.0
8.0  94.2 40.0
7.0  94.2 40.0
6.0  95.0 40.0
5.0  95.2 40.0
3.30 95.78 40.0
2.98 94.45 10.0"""
    stringio = io.StringIO(sm_data2)
    jdict = text_to_json(stringio, new_format=False)
    coords = [
        [94.45, 2.98, 10.0],
        [94.0, 5.0, 10.0],
        [93.5, 6.0, 10.0],
        [93.2, 7.0, 10.0],
        [93.1, 8.0, 10.0],
        [92.5, 10.0, 10.0],
        [92.5, 12.0, 10.0],
        [93.5, 12.0, 40.0],
        [93.5, 10.0, 40.0],
        [94.2, 8.0, 40.0],
        [94.2, 7.0, 40.0],
        [95.0, 6.0, 40.0],
        [95.2, 5.0, 40.0],
        [95.78, 3.30, 40.0],
        [94.45, 2.98, 10.0],
    ]
    assert len(jdict["features"][0]["geometry"]["coordinates"]) == 1
    assert jdict["features"][0]["geometry"]["coordinates"][0][0] == coords


def test_rupture_from_dict():
    # Grab an EdgeRupture
    origin = Origin(
        {
            "id": "test",
            "lat": 0,
            "lon": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    file = os.path.join(homedir, "rupture_data/cascadia.json")
    rup_original = get_rupture(origin, file)
    d = rup_original._geojson
    rup_from_dict = rupture_from_dict(d)
    assert rup_from_dict._mesh_dx == 0.5

    # Specify mesh_dx
    rup_original = get_rupture(origin, file, mesh_dx=1.0)
    d = rup_original._geojson
    rup_from_dict = rupture_from_dict(d)
    assert rup_from_dict._mesh_dx == 1.0

    # Quad rupture
    file = os.path.join(homedir, "rupture_data/izmit.json")
    rup_original = get_rupture(origin, file)
    d = rup_original._geojson
    rup_from_dict = rupture_from_dict(d)
    assert rup_from_dict.getArea() == rup_original.getArea()
    # Note, there's a bit of an inconsistency highlighted here because
    # magnitude has key 'magnitude' in the izmit file, but 'mag' in
    # the origin and both get retained.

    # Point rupture from json
    file = os.path.join(homedir, "rupture_data/point.json")
    rup = get_rupture(origin, file)
    assert rup.lats == 0
    assert rup.lons == 0

    # Point rupture
    origin = Origin(
        {
            "id": "test",
            "lon": -122.5,
            "lat": 37.3,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    rup_original = get_rupture(origin)
    d = rup_original._geojson
    rup_from_dict = rupture_from_dict(d)
    assert rup_from_dict.lats == 37.3
    assert rup_from_dict.lons == -122.5

    assert rup_original.getLength() == 44.668359215096281
    assert rup_original.getWidth() == 16.982436524617452
    assert rup_original.getArea() == 758.57757502918355
    assert rup_original.getStrike() == constants.DEFAULT_STRIKE
    assert rup_original.getDip() == constants.DEFAULT_DIP
    assert rup_original.getDepthToTop() == 0.0
    assert rup_original.getQuadrilaterals() is None
    assert rup_original.depths == 5.0
    # No mech, no tectonic region
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    # Various combinations of mech and tectonic region...
    rup_original._origin._tectonic_region = "Active Shallow Crust"
    rup_original._origin.mech = "ALL"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "RS"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "NM"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "SS"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin._tectonic_region = "Stable Shallow Crust"
    rup_original._origin.mech = "ALL"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "RS"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "NM"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin.mech = "SS"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])
    rup_original._origin._tectonic_region = "Somewhere Else"
    rup_original._origin.mech = "ALL"
    rjb = rup_original.computeRjb(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    rrup = rup_original.computeRrup(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    if do_tests is True:
        np.testing.assert_allclose([rjb[0], rrup[0]], [55.466511, 55.691416])
    else:
        print(rjb[0], rrup[0])

    # This is just zeroes now, so there's not much to check
    gc2 = rup_original.computeGC2(
        np.array([-122.0]), np.array([37.0]), np.array([0.0])
    )
    assert gc2["rx"][0] == 0


def test_EdgeRupture():
    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    file = os.path.join(homedir, "rupture_data/cascadia.json")
    rup = get_rupture(origin, file)
    np.testing.assert_allclose(rup.getArea(), 105635.92827547337)

    # Force read Northridge as EdgeRupture
    file = os.path.join(homedir, "rupture_data/northridge_fault.txt")
    d = text_to_json(file, new_format=True)
    rupt = EdgeRupture(d, origin)
    strike = rupt.getStrike()
    np.testing.assert_allclose(strike, 121.97, atol=0.01)
    dip = rupt.getDip()
    np.testing.assert_allclose(dip, 40.12, atol=0.01)
    L = rupt.getLength()
    np.testing.assert_allclose(L, 17.99, atol=0.01)
    W = rupt.getWidth()
    np.testing.assert_allclose(W, 23.92, atol=0.01)
    ztor = rupt.getDepthToTop()
    np.testing.assert_allclose(ztor, 5, atol=0.01)

    # And again for the same vertices but reversed order
    file = os.path.join(homedir, "rupture_data/northridge_fixed_fault.txt")
    d = text_to_json(file, new_format=True)
    rupt = EdgeRupture(d, origin)
    strike = rupt.getStrike()
    np.testing.assert_allclose(strike, 121.97, atol=0.01)
    dip = rupt.getDip()
    np.testing.assert_allclose(dip, 40.12, atol=0.01)
    L = rupt.getLength()
    np.testing.assert_allclose(L, 17.99, atol=0.01)
    W = rupt.getWidth()
    np.testing.assert_allclose(W, 23.92, atol=0.01)
    ztor = rupt.getDepthToTop()
    np.testing.assert_allclose(ztor, 5, atol=0.01)

    # Test for fromArrays method
    toplats = np.array([37.0, 38.0])
    toplons = np.array([-120.0, -120.0])
    topdeps = np.array([0.0, 0.0])
    botlats = copy.copy(toplats)
    botlons = copy.copy(toplons)
    botdeps = np.array([10.0, 10.0])
    erup = EdgeRupture.fromArrays(
        toplons, toplats, topdeps, botlons, botlats, botdeps, origin
    )

    qrup = QuadRupture.fromVertices(
        [toplons[0]],
        [toplats[0]],
        [topdeps[0]],
        [toplons[1]],
        [toplats[1]],
        [topdeps[1]],
        [botlons[1]],
        [botlats[1]],
        [botdeps[1]],
        [botlons[0]],
        [botlats[0]],
        [botdeps[0]],
        origin,
    )
    np.testing.assert_allclose(erup.getArea(), 1108.9414759967776)
    np.testing.assert_allclose(erup.getDepthToTop(), 0)
    np.testing.assert_allclose(erup.getLength(), 111.19492664455889)
    np.testing.assert_allclose(
        erup.lats, np.array([37.0, 38.0, 38.0, 37.0, 37.0, np.nan])
    )
    np.testing.assert_allclose(
        erup.lons, np.array([-120.0, -120.0, -120.0, -120.0, -120.0, np.nan])
    )
    np.testing.assert_allclose(
        erup.depths, np.array([0.0, 0.0, 10.0, 10.0, 0.0, np.nan])
    )
    np.testing.assert_allclose(erup._getGroupIndex(), np.array([0.0]))
    quads = erup.getQuadrilaterals()
    np.testing.assert_allclose(quads[0][0].x, -120.0)

    # Need to also test the distances with EdgeRupture
    lons = np.linspace(-120.1, -121.0, 10)
    lats = np.linspace(37.0, 38, 10)
    deps = np.zeros_like(lons)
    rrup1 = qrup.computeRrup(lons, lats, deps)
    rrup2 = erup.computeRrup(lons, lats, deps)
    np.testing.assert_allclose(rrup1, rrup2, atol=2e-2)
    rjb1 = qrup.computeRjb(lons, lats, deps)
    rjb2 = erup.computeRjb(lons, lats, deps)
    np.testing.assert_allclose(rjb1, rjb2, atol=2e-2)
    gc2 = erup.computeGC2(lons, lats, deps)
    targetRy0 = np.array(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.67335931]
    )
    targetRx = np.array(
        [
            -8.88024949,
            -17.73390996,
            -26.56167797,
            -35.3634266,
            -44.13902929,
            -52.88835984,
            -61.61129242,
            -70.30770154,
            -78.97746209,
            -87.6204493,
        ]
    )
    np.testing.assert_allclose(gc2["ry0"], targetRy0)
    np.testing.assert_allclose(gc2["rx"], targetRx)


def test_QuadRupture():
    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    # First with json file
    file = os.path.join(homedir, "rupture_data/izmit.json")
    rupj = get_rupture(origin, file)
    # Then with text file:
    file = os.path.join(homedir, "rupture_data/Barkaetal02_fault.txt")
    rupt = get_rupture(origin, file)

    np.testing.assert_allclose(rupj.lats, rupt.lats, atol=1e-5)
    np.testing.assert_allclose(rupj.lons, rupt.lons, atol=1e-5)
    np.testing.assert_allclose(rupj._depth, rupt._depth, atol=1e-5)
    np.testing.assert_allclose(rupt.getArea(), 2391.2822653900268, atol=1e-5)

    target = np.array(
        [
            29.51528,
            29.3376,
            29.3376,
            29.51528005,
            29.51528,
            np.nan,
            29.87519,
            29.61152,
            29.61152,
            29.87519021,
            29.87519,
            np.nan,
            30.11126,
            29.88662,
            30.11126,
            30.11126,
            29.88662,
            30.11126,
            30.11126,
            np.nan,
            30.4654,
            30.30494,
            30.4654,
            30.4654,
            30.30494,
            30.4654,
            30.4654,
            np.nan,
            30.63731,
            30.57658,
            30.57658,
            30.63731011,
            30.63731,
            np.nan,
            30.93655,
            30.729,
            30.729,
            30.93655103,
            30.93655,
            np.nan,
            31.01799,
            30.94688,
            30.94688,
            31.0179905,
            31.01799,
            np.nan,
        ]
    )
    np.testing.assert_allclose(rupj.lons, target, atol=1e-5)
    target = np.array(
        [
            40.72733,
            40.70985,
            40.71185,
            40.72932969,
            40.72733,
            np.nan,
            40.74903,
            40.70513,
            40.70713,
            40.75102924,
            40.74903,
            np.nan,
            40.72336,
            40.72582,
            40.72336,
            40.72536,
            40.72782,
            40.72536004,
            40.72336,
            np.nan,
            40.71081,
            40.7121,
            40.71081,
            40.71281,
            40.7141,
            40.71281002,
            40.71081,
            np.nan,
            40.70068,
            40.71621,
            40.71821,
            40.70268025,
            40.70068,
            np.nan,
            40.79654,
            40.69947,
            40.70147,
            40.79853872,
            40.79654,
            np.nan,
            40.84501,
            40.80199,
            40.80399,
            40.84700952,
            40.84501,
            np.nan,
        ]
    )
    np.testing.assert_allclose(rupj.lats, target, atol=1e-5)
    target = np.array(
        [
            -0.00000000e00,
            -0.00000000e00,
            2.00000000e01,
            1.99999325e01,
            -0.00000000e00,
            np.nan,
            -9.31322575e-13,
            -0.00000000e00,
            2.00000000e01,
            1.99998304e01,
            -9.31322575e-13,
            np.nan,
            9.31322575e-13,
            -0.00000000e00,
            9.31322575e-13,
            2.00000000e01,
            2.00000000e01,
            2.00000095e01,
            9.31322575e-13,
            np.nan,
            -0.00000000e00,
            -0.00000000e00,
            -0.00000000e00,
            2.00000000e01,
            2.00000000e01,
            2.00000050e01,
            -0.00000000e00,
            np.nan,
            -0.00000000e00,
            -0.00000000e00,
            2.00000000e01,
            2.00000600e01,
            -0.00000000e00,
            np.nan,
            -0.00000000e00,
            -0.00000000e00,
            2.00000000e01,
            1.99996249e01,
            -0.00000000e00,
            np.nan,
            -0.00000000e00,
            -0.00000000e00,
            2.00000000e01,
            1.99998338e01,
            -0.00000000e00,
            np.nan,
        ]
    )
    np.testing.assert_allclose(rupj.depths, target, atol=1e-5)


def test_slip():
    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    # Make a rupture
    lat0 = np.array([34.1])
    lon0 = np.array([-118.2])
    lat1 = np.array([34.2])
    lon1 = np.array([-118.15])
    z = np.array([1.0])
    W = np.array([3.0])
    dip = np.array([30.0])
    rup = QuadRupture.fromTrace(lon0, lat0, lon1, lat1, z, W, dip, origin)

    slp = get_quad_slip(rup.getQuadrilaterals()[0], 30).getArray()
    slpd = np.array([0.80816457, 0.25350787, 0.53160491])
    np.testing.assert_allclose(slp, slpd)

    slp = get_quad_strike_vector(rup.getQuadrilaterals()[0]).getArray()
    slpd = np.array([0.58311969, 0.27569625, 0.76417472])
    np.testing.assert_allclose(slp, slpd)

    slp = get_quad_down_dip_vector(rup.getQuadrilaterals()[0]).getArray()
    slpd = np.array([0.81219873, -0.17763484, -0.55567895])
    np.testing.assert_allclose(slp, slpd)

    slp = get_local_unit_slip_vector(22, 30, 86).getArray()
    slpd = np.array([0.82714003, 0.38830563, 0.49878203])
    np.testing.assert_allclose(slp, slpd)

    slp = get_local_unit_slip_vector_DS(22, 30, -86).getArray()
    slpd = np.array([-0.80100879, -0.32362856, -0.49878203])
    np.testing.assert_allclose(slp, slpd)

    slp = get_local_unit_slip_vector_SS(22, 80, 5).getArray()
    slpd = np.array([0.3731811, 0.92365564, 0.0])
    np.testing.assert_allclose(slp, slpd)

    mech = rake_to_mech(-160)
    assert mech == "SS"
    mech = rake_to_mech(0)
    assert mech == "SS"
    mech = rake_to_mech(160)
    assert mech == "SS"
    mech = rake_to_mech(-80)
    assert mech == "NM"
    mech = rake_to_mech(80)
    assert mech == "RS"


def test_northridge():
    rupture_text = """# Source: Wald, D. J., T. H. Heaton, and K. W. Hudnut \
(1996). The Slip History of the 1994 Northridge, California, Earthquake \
Determined from Strong-Motion, Teleseismic, GPS, and Leveling Data, Bull. \
Seism. Soc. Am. 86, S49-S70.
    -118.421 34.315  5.000
    -118.587 34.401  5.000
    -118.693 34.261 20.427
    -118.527 34.175 20.427
    -118.421 34.315 5.000
    """  # noqa

    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    cbuf = io.StringIO(rupture_text)
    rupture = get_rupture(origin, cbuf)
    strike = rupture.getStrike()
    np.testing.assert_allclose(strike, 122.06, atol=0.01)
    dip = rupture.getDip()
    np.testing.assert_allclose(dip, 40.21, atol=0.01)
    L = rupture.getLength()
    np.testing.assert_allclose(L, 17.99, atol=0.01)
    W = rupture.getWidth()
    np.testing.assert_allclose(W, 23.94, atol=0.01)
    nq = rupture.getNumQuads()
    np.testing.assert_allclose(nq, 1)
    ng = rupture.getNumGroups()
    np.testing.assert_allclose(ng, 1)
    nd = rupture.getDeps()
    np.testing.assert_allclose(nd, [5.0, 5.0, 20.427, 20.427, np.nan])
    sind = rupture._getGroupIndex()
    np.testing.assert_allclose(sind, [0])
    ztor = rupture.getDepthToTop()
    np.testing.assert_allclose(ztor, 5, atol=0.01)
    itl = rupture.getIndividualTopLengths()
    np.testing.assert_allclose(itl, 17.99, atol=0.01)
    iw = rupture.getIndividualWidths()
    np.testing.assert_allclose(iw, 23.94, atol=0.01)
    lats = rupture.lats
    lats_d = np.array([34.401, 34.315, 34.175, 34.261, 34.401, np.nan])
    np.testing.assert_allclose(lats, lats_d, atol=0.01)
    lons = rupture.lons
    lons_d = np.array(
        [-118.587, -118.421, -118.527, -118.693, -118.587, np.nan]
    )
    np.testing.assert_allclose(lons, lons_d, atol=0.01)
    ln, lt, de = rupture.getRuptureAsArrays()
    np.testing.assert_allclose(
        ln,
        np.array([-118.421, -118.587, -118.693, -118.527, np.nan]),
        atol=0.01,
    )
    np.testing.assert_allclose(
        lt, np.array([34.315, 34.401, 34.261, 34.175, np.nan]), atol=0.01
    )
    np.testing.assert_allclose(de, [5.0, 5.0, 20.427, 20.427, np.nan])
    mesh = rupture.getRuptureAsMesh()
    np.testing.assert_allclose(
        mesh.lons, [-118.421, -118.587, -118.693, -118.527, np.nan]
    )
    np.testing.assert_allclose(
        mesh.lats, [34.315, 34.401, 34.261, 34.175, np.nan]
    )
    np.testing.assert_allclose(mesh.depths, [5.0, 5.0, 20.427, 20.427, np.nan])


def test_parse_complicated_rupture():
    rupture_text = """# SOURCE: Barka, A., H. S. Akyz, E. Altunel, G. Sunal, \
Z. Akir, A. Dikbas, B. Yerli, R. Armijo, B. Meyer, J. B. d. Chabalier, \
T. Rockwell, J. R. Dolan, R. Hartleb, T. Dawson, S. Christofferson, \
A. Tucker, T. Fumal, R. Langridge, H. Stenner, W. Lettis, J. Bachhuber, \
and W. Page (2002). The Surface Rupture and Slip Distribution of the \
17 August 1999 Izmit Earthquake (M 7.4), North Anatolian Fault, Bull. \
Seism. Soc. Am. 92, 43-60.
    29.33760 40.70985 0
    29.51528 40.72733 0
    29.51528 40.72933 20
    29.33760 40.71185 20
    29.33760 40.70985 0
    >
    29.61152 40.70513 0
    29.87519 40.74903 0
    29.87519 40.75103 20
    29.61152 40.70713 20
    29.61152 40.70513 0
    >
    29.88662 40.72582 0
    30.11126 40.72336 0
    30.19265 40.73432 0
    30.19265 40.73632 20
    30.11126 40.72536 20
    29.88662 40.72782 20
    29.88662 40.72582 0
    >
    30.30494 40.71210 0
    30.46540 40.71081 0
    30.56511 40.70739 0
    30.56511 40.70939 20
    30.46540 40.71281 20
    30.30494 40.71410 20
    30.30494 40.71210 0
    >
    30.57658 40.71621 0
    30.63731 40.70068 0
    30.63731 40.70268 20
    30.57658 40.71821 20
    30.57658 40.71621 0
    >
    30.72900 40.69947 0
    30.93655 40.79654 0
    30.93655 40.79854 20
    30.72900 40.70147 20
    30.72900 40.69947 0
    >
    30.94688 40.80199 0
    31.01799 40.84501 0
    31.01799 40.84701 20
    30.94688 40.80399 20
    30.94688 40.80199 0"""  # noqa

    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )
    cbuf = io.StringIO(rupture_text)
    rupture = get_rupture(origin, cbuf)
    strike = rupture.getStrike()
    np.testing.assert_allclose(strike, -100.46, atol=0.01)
    dip = rupture.getDip()
    np.testing.assert_allclose(dip, 89.40, atol=0.01)
    L = rupture.getLength()
    np.testing.assert_allclose(L, 119.56, atol=0.01)
    W = rupture.getWidth()
    np.testing.assert_allclose(W, 20.0, atol=0.01)
    nq = rupture.getNumQuads()
    np.testing.assert_allclose(nq, 9)
    ng = rupture.getNumGroups()
    np.testing.assert_allclose(ng, 7)
    sind = rupture._getGroupIndex()
    np.testing.assert_allclose(sind, [0, 1, 2, 2, 3, 3, 4, 5, 6])
    ztor = rupture.getDepthToTop()
    np.testing.assert_allclose(ztor, 0, atol=0.01)
    itl = rupture.getIndividualTopLengths()
    itl_d = np.array(
        [
            15.13750778,
            22.80237887,
            18.98053425,
            6.98263853,
            13.55978731,
            8.43444811,
            5.41399812,
            20.57788056,
            7.66869463,
        ]
    )
    np.testing.assert_allclose(itl, itl_d, atol=0.01)
    iw = rupture.getIndividualWidths()
    iw_d = np.array(
        [
            20.00122876,
            20.00122608,
            20.00120173,
            20.00121028,
            20.00121513,
            20.00121568,
            20.00107293,
            20.00105498,
            20.00083348,
        ]
    )
    np.testing.assert_allclose(iw, iw_d, atol=0.01)
    lats = rupture.lats
    lats_d = np.array(
        [
            40.72733,
            40.70985,
            40.71185,
            40.72932969,
            40.72733,
            np.nan,
            40.74903,
            40.70513,
            40.70713,
            40.75102924,
            40.74903,
            np.nan,
            40.72336,
            40.72582,
            40.72336,
            40.72536,
            40.72782,
            40.72536004,
            40.72336,
            np.nan,
            40.71081,
            40.7121,
            40.71081,
            40.71281,
            40.7141,
            40.71281002,
            40.71081,
            np.nan,
            40.70068,
            40.71621,
            40.71821,
            40.70268025,
            40.70068,
            np.nan,
            40.79654,
            40.69947,
            40.70147,
            40.79853872,
            40.79654,
            np.nan,
            40.84501,
            40.80199,
            40.80399,
            40.84700952,
            40.84501,
            np.nan,
        ]
    )
    np.testing.assert_allclose(lats, lats_d, atol=0.001)
    lons = rupture.lons
    lons_d = np.array(
        [
            29.51528,
            29.3376,
            29.3376,
            29.51528005,
            29.51528,
            np.nan,
            29.87519,
            29.61152,
            29.61152,
            29.87519021,
            29.87519,
            np.nan,
            30.11126,
            29.88662,
            30.11126,
            30.11126,
            29.88662,
            30.11126,
            30.11126,
            np.nan,
            30.4654,
            30.30494,
            30.4654,
            30.4654,
            30.30494,
            30.4654,
            30.4654,
            np.nan,
            30.63731,
            30.57658,
            30.57658,
            30.63731011,
            30.63731,
            np.nan,
            30.93655,
            30.729,
            30.729,
            30.93655103,
            30.93655,
            np.nan,
            31.01799,
            30.94688,
            30.94688,
            31.0179905,
            31.01799,
            np.nan,
        ]
    )

    np.testing.assert_allclose(lons, lons_d, atol=0.001)


def test_incorrect():
    # Number of points in polyon is even
    rupture_text = """# Source: Ji, C., D. V. Helmberger, D. J. Wald, and \
K.-F. Ma (2003). Slip history and dynamic implications of the 1999 Chi-Chi, \
Taiwan, earthquake, J. Geophys. Res. 108, 2412, doi:10.1029/2002JB001764.
    120.72300 24.27980 	0
    121.00000 24.05000	17
    121.09300 24.07190	17
    121.04300 24.33120	17
    121.04300 24.33120	17
    120.72300 24.27980	0
    >
    120.72300 24.27980	0
    120.68000 23.70000	0
    120.97200 23.60400	17
    121.00000 24.05000	17
    120.72300 24.27980	0
    >
    120.97200 23.60400	17
    120.68000 23.70000	0
    120.58600 23.58850	0
    120.78900 23.40240	17
    120.97200 23.60400	17"""  # noqa

    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf)

    # Top points must be first
    rupture_text = """# Test
    120.72300 24.27980 	0
    121.00000 24.05000	17
    121.09300 24.07190	17
    121.04300 24.33120	17
    120.72300 24.27980	0"""  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf)

    # Wrong order of lat/lon
    rupture_text = """# Test
    -118.421 34.315  5.000
    -118.587 34.401  5.000
    -118.693 34.261 20.427
    -118.527 34.175 20.427
    -118.421 34.315 5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=False)

    # Wrong order of lat/lon
    rupture_text = """# Test
    34.315 -118.421  5.000
    34.401 -118.587  5.000
    34.261 -118.693 20.427
    34.175 -118.527 20.427
    34.315 -118.421  5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=True)

    # Unclosed segments
    rupture_text = """# Test
    34.315 -118.421  5.000
    34.401 -118.587  5.000
    34.261 -118.693 20.427
    34.175 -118.527 20.427
    34.315 -118.6    5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=False)

    # incorrect delimiter
    rupture_text = """#Test
    34.315;-118.421;5.000
    34.401;-118.587;5.000
    34.261;-118.693;20.427
    34.175;-118.527;20.427
    34.315;-118.421;5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=False)

    # incorrect delimiter, new format
    rupture_text = """#Test
    34.315;-118.421;5.000
    34.401;-118.587;5.000
    34.261;-118.693;20.427
    34.175;-118.527;20.427
    34.315;-118.421;5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=True)

    # Not 3 columns
    rupture_text = """#Test
    34.315 -118.421;5.000
    34.401 -118.587;5.000
    34.261 -118.693;20.427
    34.175 -118.527;20.427
    34.315 -118.421;5.000
    """  # noqa
    cbuf = io.StringIO(rupture_text)
    with pytest.raises(Exception):
        get_rupture(origin, cbuf, new_format=False)

    # Json incorrect
    test = {
        "metadata": {
            "id": "test",
            "mag": 7.0,
            "lon": 0,
            "mech": "ALL",
            "depth": 5.0,
            "time": "2018-07-02T22:50:03Z",
            "netid": "us",
            "rake": 0.0,
            "lat": 0,
            "network": "",
            "locstring": "",
            "reference": "Test",
        },
        "features": [
            {
                "type": "Feature",
                "geometry": {
                    "coordinates": [
                        [
                            [
                                [-118.421, 34.315, 5.0],
                                [-118.587, 34.401, 5.0],
                                [-118.693, 34.261, 20.427],
                                [-118.527, 34.175, 20.427],
                                [-118.421, 34.315, 5.0],
                            ]
                        ]
                    ],
                    "type": "MultiPolygon",
                },
                "properties": {"rupture type": "rupture extent"},
            }
        ],
        "type": "FeatureCollection",
    }

    # incorrect type
    test_incorrect = copy.deepcopy(test)
    test_incorrect["type"] = "Feature"
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))

    # Incorrect number of features
    test_incorrect = copy.deepcopy(test)
    test_incorrect["features"].append(["wrong"])
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))

    # no reference
    test_incorrect = copy.deepcopy(test)
    test_incorrect["metadata"].pop("reference", None)
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))

    # incorrect feature type
    test_incorrect = copy.deepcopy(test)
    test_incorrect["features"][0]["type"] = "fred"
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))

    # incorrect feature geometry type
    test_incorrect = copy.deepcopy(test)
    test_incorrect["features"][0]["geometry"]["type"] = "fred"
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))

    # no coordinates
    test_incorrect = copy.deepcopy(test)
    test_incorrect["features"][0]["geometry"].pop("coordinates", None)
    with pytest.raises(Exception) as e:
        validate_json(test_incorrect)
    print(str(e))


def test_fromTrace():
    xp0 = [0.0]
    xp1 = [0.0]
    yp0 = [0.0]
    yp1 = [0.05]
    zp = [0.0]
    widths = [10.0]
    dips = [45.0]

    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": -121.81529,
            "lat": 37.73707,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )

    # TODO: These write tests exercise code, but we don't check the results
    rupture = QuadRupture.fromTrace(
        xp0,
        yp0,
        xp1,
        yp1,
        zp,
        widths,
        dips,
        origin,
        reference="From J Smith, (personal communication)",
    )
    fstr = io.StringIO()
    rupture.writeTextFile(fstr)

    tfile = tempfile.NamedTemporaryFile()
    tname = tfile.name
    tfile.close()
    rupture.writeTextFile(tname)
    os.remove(tname)

    tfile = tempfile.NamedTemporaryFile()
    tname = tfile.name
    tfile.close()
    rupture.writeGeoJson(tname)
    os.remove(tname)

    xp0 = [-121.81529, -121.82298]
    xp1 = [-121.82298, -121.83068]
    yp0 = [37.73707, 37.74233]
    yp1 = [37.74233, 37.74758]
    zp = [10, 15]
    widths = [15.0, 20.0]
    dips = [30.0, 45.0]
    rupture = QuadRupture.fromTrace(
        xp0,
        yp0,
        xp1,
        yp1,
        zp,
        widths,
        dips,
        origin,
        reference="From J Smith, (personal communication)",
    )

    assert rupture.getReference() == "From J Smith, (personal communication)"
    rorigin = rupture.getOrigin()
    assert rorigin.id == origin.id
    assert rorigin.mag == origin.mag
    assert rorigin.depth == origin.depth

    rx = rupture.getRuptureContext([])
    np.testing.assert_allclose(
        [rx.strike, rx.dip, rx.ztor, rx.width],
        [
            -49.183708644954905,
            37.638322472702534,
            9.999999999371358,
            17.47024205615428,
        ],
    )

    rhyp = rupture.computeRhyp(
        np.array([-121.5]), np.array([37.0]), np.array([0])
    )
    repi = rupture.computeRepi(
        np.array([-121.5]), np.array([37.0]), np.array([0])
    )
    np.testing.assert_allclose([rhyp[0], repi[0]], [86.709236, 86.564956])


def test_with_quakeml():
    np1 = NodalPlane(strike=259, dip=74, rake=10)
    np2 = NodalPlane(strike=166, dip=80, rake=164)
    nodal_planes = NodalPlanes(nodal_plane_1=np1, nodal_plane_2=np2)
    taxis = Axis(plunge=40, azimuth=70)
    naxis = Axis(plunge=50, azimuth=80)
    paxis = Axis(plunge=60, azimuth=90)
    paxes = PrincipalAxes(t_axis=taxis, n_axis=naxis, p_axis=paxis)
    focal = FocalMechanism(nodal_planes=nodal_planes, principal_axes=paxes)
    event = Event(focal_mechanisms=[focal])
    catalog = Catalog(events=[event])
    event_text = """<shakemap-data code_version="4.0" map_version="1">
<earthquake id="us2000cmy3" lat="56.046" lon="-149.073" mag="7.9"
time="2018-01-23T09:31:42Z"
depth="25.00" locstring="280km SE of Kodiak, Alaska" netid="us" network=""/>
</shakemap-data>"""
    try:
        tempdir = tempfile.mkdtemp()
        xmlfile = os.path.join(tempdir, "quakeml.xml")
        catalog.write(xmlfile, format="QUAKEML")
        eventfile = os.path.join(tempdir, "event.xml")
        f = open(eventfile, "wt")
        f.write(event_text)
        f.close()
        params = read_moment_quakeml(xmlfile)
        assert params["moment"]["NP1"]["strike"] == 259.0
        assert params["moment"]["NP1"]["dip"] == 74.0
        assert params["moment"]["NP1"]["rake"] == 10.0
        assert params["moment"]["NP2"]["strike"] == 166.0
        assert params["moment"]["NP2"]["dip"] == 80.0
        assert params["moment"]["NP2"]["rake"] == 164.0
        origin = Origin.fromFile(eventfile, momentfile=xmlfile)
        assert origin.mag == 7.9
        assert origin.lat == 56.046
        assert origin.lon == -149.073
        assert origin.id == "us2000cmy3"
    except Exception:
        assert False
    finally:
        shutil.rmtree(tempdir)


def test_fromOrientation():
    py = [0, 0.5]
    px = [0, 0.5]
    pz = [10, 20]
    dx = [5, 7]
    dy = [8, 5]
    width = [10, 40]
    length = [20, 50]
    strike = [0, 90]
    dip = [30, 20]

    # Rupture requires an origin even when not used:
    origin = Origin(
        {
            "id": "test",
            "lon": 0,
            "lat": 0,
            "depth": 5.0,
            "mag": 7.0,
            "netid": "us",
            "network": "",
            "locstring": "",
            "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
        }
    )
    rupture = QuadRupture.fromOrientation(
        px, py, pz, dx, dy, length, width, strike, dip, origin
    )
    p1 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][0][0]
    p2 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][0][1]
    p3 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][0][2]
    p4 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][0][3]
    p5 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][1][0]
    p6 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][1][1]
    p7 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][1][2]
    p8 = rupture._geojson["features"][0]["geometry"]["coordinates"][0][1][3]

    # Check depths
    np.testing.assert_allclose(p1[2], 6)
    np.testing.assert_allclose(p2[2], 6)
    np.testing.assert_allclose(p3[2], 11)
    np.testing.assert_allclose(p4[2], 11)
    np.testing.assert_allclose(p5[2], 18.2898992834)
    np.testing.assert_allclose(p6[2], 18.2898992834)
    np.testing.assert_allclose(p7[2], 31.9707050164)
    np.testing.assert_allclose(p8[2], 31.9707050164)

    # Exception raised if no origin
    with pytest.raises(Exception) as a:
        rupture = QuadRupture.fromOrientation(
            px, py, pz, dx, dy, length, width, strike, dip, None
        )
    print(str(a))

    # Exception raised if different lengths of arrays
    with pytest.raises(Exception) as a:
        py = [0, 2]
        px = [0]
        pz = [10]
        dx = [5]
        dy = [8]
        width = [10]
        length = [20]
        strike = [0]
        dip = [30]

        origin = Origin(
            {
                "id": "test",
                "lon": 0,
                "lat": 0,
                "depth": 5.0,
                "mag": 7.0,
                "netid": "us",
                "network": "",
                "locstring": "",
                "time": HistoricTime.fromtimestamp(time.time(), datetime.UTC),
            }
        )
        rupture = QuadRupture.fromOrientation(
            px, py, pz, dx, dy, length, width, strike, dip, origin
        )
    print(str(a))


if __name__ == "__main__":
    test_no_origin()
    test_text_to_json()
    test_rupture_from_dict()
    test_EdgeRupture()
    test_QuadRupture()
    test_slip()
    test_northridge()
    test_parse_complicated_rupture()
    test_incorrect()
    test_fromTrace()
    test_with_quakeml()
    test_fromOrientation()



#!/usr/bin/env python

# stdlib imports
import io
import os
import tempfile

# third party
import numpy as np
import pytest
from esi_utils_rupture.origin import Origin
from esi_utils_rupture.origin import write_event_file
from esi_utils_time.ancient_time import HistoricTime
from esi_utils_rupture import constants


def test_origin():
    fault_text = """30.979788       103.454422      1
31.691615       104.419160      1
31.723569       104.374760      1
32.532213       105.220821      1
32.641450       105.135050      20
31.846790       104.246202      20
31.942158       104.205286      20
31.290105       103.284388      20
30.979788       103.454422      1"""
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008ryan" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" netid="us" network=""
locstring="EASTERN SICHUAN, CHINA"
mech="" />"""
    source_text = "mech=RS"
    ffile = io.StringIO(fault_text)  # noqa
    efile = io.StringIO(event_text)
    sfile = io.StringIO(source_text)
    origin = Origin.fromFile(efile, sourcefile=sfile)

    testdict = {
        "mag": 7.9,
        "id": "2008ryan",
        "locstring": "EASTERN SICHUAN, CHINA",
        "mech": "RS",
        "lon": 103.3639,
        "lat": 30.9858,
        "depth": 19.0,
    }
    for key in testdict.keys():
        value = eval(f"origin.{key}")
        if type(value) is str:
            assert testdict[key] == value
        if type(value) is float:
            np.testing.assert_almost_equal(testdict[key], value)

    assert origin.mech == "RS"
    origin.setMechanism("SS")
    assert origin.mech == "SS"
    origin.setMechanism("SS", rake=10)
    assert origin.mech == "SS"
    assert origin.rake == 10.0
    assert origin.dip == 90.0
    origin.setMechanism("SS", rake=-350)
    assert origin.rake == 10.0
    origin.setMechanism("SS", rake=370)
    assert origin.rake == 10.0
    with pytest.raises(Exception) as a:
        origin.setMechanism("SS", dip=100)
    with pytest.raises(Exception) as a:
        origin.setMechanism("Strike slip")
    # Rake too large
    with pytest.raises(Exception) as a:
        origin.setMechanism("SS", rake=1111)
    # Lat is greater than 90
    with pytest.raises(Exception) as a:
        event_text = """<?xml version="1.0" encoding="US-ASCII"
standalone="yes"?> <earthquake id="2008" lat="91.9858" lon="103.3639"
mag="7.9" time="2008-05-12T06:28:01Z"
timezone="GMT" depth="19.0" locstring="EASTERN SICHUAN, CHINA"
created="1211173621" otime="1210573681" mech="" netid="us" network=""/>"""
        efile = io.StringIO(event_text)
        origin = Origin.fromFile(efile)
    # Lon is greater than 180
    with pytest.raises(Exception) as a:
        event_text = """<?xml version="1.0" encoding="US-ASCII"
standalone="yes"?> <earthquake id="2008" lat="31.9858" lon="183.3639"
mag="7.9" time="2008-05-12T06:28:01Z"
timezone="GMT" depth="19.0" locstring="EASTERN SICHUAN, CHINA"
created="1211173621" otime="1210573681" type="" netid="us" network=""/>"""
        efile = io.StringIO(event_text)
        origin = Origin.fromFile(efile)
    # No event id
    with pytest.raises(Exception) as a:
        event_text = """<?xml version="1.0" encoding="US-ASCII"
standalone="yes"?> <earthquake lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
timezone="GMT" depth="19.0" locstring="EASTERN SICHUAN, CHINA"
created="1211173621" otime="1210573681" type="" netid="us" network=""/>"""
        efile = io.StringIO(event_text)
        origin = Origin.fromFile(efile)
    # Put mech in event keys
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" locstring="EASTERN SICHUAN, CHINA" created="1211173621"
otime="1210573681" type="" mech="SS" netid="us" network=""/>"""
    efile = io.StringIO(event_text)
    origin = Origin.fromFile(efile)
    assert origin.mech == "SS"
    # Empty mech
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" locstring="EASTERN SICHUAN, CHINA" created="1211173621"
otime="1210573681" type="" mech="" netid="us" network=""/>"""
    efile = io.StringIO(event_text)
    origin = Origin.fromFile(efile)
    assert origin.mech == "ALL"
    # Mech not acceptable value
    with pytest.raises(Exception) as a:  # noqa
        event_text = """<?xml version="1.0" encoding="US-ASCII"
standalone="yes"?> <earthquake id="2008" lat="31.9858" lon="103.3639"
mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" locstring="EASTERN SICHUAN, CHINA"
created="1211173621" otime="1210573681" type="" mech="Strike slip"
netid="us" network=""/>"""
        efile = io.StringIO(event_text)
        origin = Origin.fromFile(efile)

    # Missing keys
    with pytest.raises(KeyError):
        event_text = """<?xml version="1.0" encoding="US-ASCII"
standalone="yes"?> <earthquake id="2008"
mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" locstring="EASTERN SICHUAN, CHINA"
created="1211173621" otime="1210573681" type=""
network=""/>"""
        efile = io.StringIO(event_text)
        origin = Origin.fromFile(efile)

    # Use "type" instead of "mech"
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008ryan" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" netid="us" network=""
locstring="EASTERN SICHUAN, CHINA"
type="RS" />"""
    efile = io.StringIO(event_text)
    origin = Origin.fromFile(efile)
    assert origin.mech == "RS"

    # No rake or mech
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008ryan" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" netid="us" network=""
locstring="EASTERN SICHUAN, CHINA"
reference="Smith, et al. (2019)"
 />"""
    efile = io.StringIO(event_text)
    origin = Origin.fromFile(efile)
    assert origin.rake == 0.0

    hypo = origin.getHypo()
    assert hypo.x == origin.lon
    assert hypo.y == origin.lat
    assert hypo.z == origin.depth

    # Write the origin to a file
    event = {}
    event["id"] = "us2000ryan"
    event["netid"] = "us"
    event["network"] = "USGS Network"
    event["lat"] = 30.9858
    event["lon"] = 103.3639
    event["depth"] = 19.0
    event["mag"] = 7.9
    event["time"] = HistoricTime.strptime("2008-05-12T06:28:01.0Z", constants.TIMEFMT)
    event["locstring"] = "EASTERN SICHUAN, CHINA"
    event["mech"] = "RS"
    event["reference"] = "Smith, et al. (2019)"
    event["productcode"] = "us2000ryan"

    tfile = tempfile.NamedTemporaryFile()
    xmlfile = tfile.name
    tfile.close()
    res = write_event_file(event, xmlfile)
    if res is not None:
        print(res)
        assert False

    origin = Origin.fromFile(xmlfile)
    os.remove(xmlfile)
    assert origin.id == event["id"]
    assert origin.netid == event["netid"]
    assert origin.network == event["network"]
    assert origin.time == event["time"]


if __name__ == "__main__":
    test_origin()



#!/usr/bin/env python

# stdlib modules

# third party imports
import numpy as np

try:
    _ = np.RankWarning  # will work on numpy < 2
except AttributeError:
    setattr(np, "RankWarning", RuntimeWarning)  # will work on numpy > 2
from openquake.hazardlib.geo.point import Point

from esi_utils_vectors.ecef import latlon2ecef
from esi_utils_vectors.vector import Vector
from esi_utils_time.ancient_time import HistoricTime
from esi_utils_rupture.base import Rupture
from esi_utils_rupture import constants
from esi_utils_rupture import utils
from esi_utils_rupture import gc2


class EdgeRupture(Rupture):
    """
    Rupture class that representst the rupture surface by specifying the top
    edge and the bottom edges. These edges do not need to be horizontal. The
    freedom to allow for non-horizontal edges (as compared to QuadRupture)
    comes at the cost of slower distance calculations. This is because the
    rupture must be discretized and then the distances are compued in a brute
    force fashion based on this mesh, which can be quite large.
    """

    def __init__(self, d, origin, mesh_dx=0.5):
        """
        Initialization of an EdgeRupture from a GeoJSON dictionary and an
        Origin.

        Args:
            d (dict): Rupture GeoJSON dictionary.
            origin (Origin): Reference to a ShakeMap Origin object.
            mesh_dx (float): Target spacing (in km) for rupture discretization;
                default is 0.5 km and it is only used if the rupture file is an
                EdgeRupture.

        Returns:
            EdgeRupture instance.

        """
        polys = d["features"][0]["geometry"]["coordinates"][0]
        n_polygons = len(polys)
        toplons = np.empty(shape=(0, 0))
        toplats = np.empty(shape=(0, 0))
        topdeps = np.empty(shape=(0, 0))
        botlons = np.empty(shape=(0, 0))
        botlats = np.empty(shape=(0, 0))
        botdeps = np.empty(shape=(0, 0))
        g_ind = 0
        group_index = []
        for i in range(n_polygons):
            p = polys[i]
            n_points = len(p)
            n_pairs = int((n_points - 1) / 2)

            p_lons = [pt[0] for pt in p][0:-1]
            p_lats = [pt[1] for pt in p][0:-1]
            p_depths = [pt[2] for pt in p][0:-1]

            tlon = np.array(p_lons[0:n_pairs])
            blon = np.array(p_lons[(n_pairs):])[::-1]
            tlat = np.array(p_lats[0:n_pairs])
            blat = np.array(p_lats[(n_pairs):])[::-1]
            tdep = np.array(p_depths[0:n_pairs])
            bdep = np.array(p_depths[(n_pairs):])[::-1]

            toplons = np.append(toplons, tlon)
            toplats = np.append(toplats, tlat)
            topdeps = np.append(topdeps, tdep)
            botlons = np.append(botlons, blon)
            botlats = np.append(botlats, blat)
            botdeps = np.append(botdeps, bdep)

            group_index.extend([g_ind] * n_pairs)
            g_ind = g_ind + 1

        reference = d["metadata"]["reference"]

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, HistoricTime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v
        d["metadata"]["mesh_dx"] = mesh_dx

        self._geojson = d

        self._toplons = np.array(toplons)
        self._toplats = np.array(toplats)
        self._topdeps = np.array(topdeps)
        self._botlons = np.array(botlons)
        self._botlats = np.array(botlats)
        self._botdeps = np.array(botdeps)
        self._origin = origin
        self._group_index = np.array(group_index)
        self._mesh_dx = mesh_dx
        self._reference = reference
        self._computeStrikeDip()

    @classmethod
    def fromArrays(
        cls,
        toplons,
        toplats,
        topdeps,
        botlons,
        botlats,
        botdeps,
        origin,
        group_index=None,
        mesh_dx=0.5,
        reference="",
    ):
        """
        Class method to initialize an EdgeRupture class from arrays of
        longitude, latitude, and depth along the top and bottom edges.

        Args:
            toplons (ndarray): Array of top edge longitudes.
            toplats (ndarray): Array of top edge latitudes.
            topdeps (ndarray): Array of top edge depths (km).
            botlons (ndarray): Array of bot edge longitudes.
            botlats (ndarray): Array of bot edge latitudes.
            botdeps (ndarray): Array of bot edge depths (km).
            origin (Origin): Reference to a ShakeMap Origin object.
            group_index (ndarray): Optional array of group index.
                If None, then assume only single group.
            mesh_dx (float): Target spacing (in km) for rupture discretization.
            reference (str): Citable reference for rupture.

        Returns:
            EdgeRupture instance.

        """
        toplons = np.array(toplons)
        toplats = np.array(toplats)
        topdeps = np.array(topdeps)
        botlons = np.array(botlons)
        botlats = np.array(botlats)
        botdeps = np.array(botdeps)
        if group_index is not None:
            group_index = np.array(group_index)
        else:
            group_index = np.zeros_like(toplons)

        coords = []
        u_groups = np.unique(group_index)
        n_groups = len(u_groups)
        for i in range(n_groups):
            ind = np.where(u_groups[i] == group_index)[0]
            lons = np.concatenate(
                [
                    toplons[ind],
                    botlons[ind][::-1],
                    toplons[ind][0].reshape((1,)),
                ]
            )
            lats = np.concatenate(
                [
                    toplats[ind],
                    botlats[ind][::-1],
                    toplats[ind][0].reshape((1,)),
                ]
            )
            deps = np.concatenate(
                [
                    topdeps[ind],
                    botdeps[ind][::-1],
                    topdeps[ind][0].reshape((1,)),
                ]
            )
            poly = []
            for lon, lat, dep in zip(lons, lats, deps):
                poly.append([lon, lat, dep])
            coords.append(poly)

        d = {
            "type": "FeatureCollection",
            "metadata": {"reference": reference},
            "features": [
                {
                    "type": "Feature",
                    "properties": {"rupture type": "rupture extent"},
                    "geometry": {
                        "type": "MultiPolygon",
                        "coordinates": [coords],
                    },
                }
            ],
        }

        # Add origin information to metadata
        odict = origin.__dict__
        for k, v in odict.items():
            if isinstance(v, HistoricTime):
                d["metadata"][k] = v.strftime(constants.TIMEFMT)
            else:
                d["metadata"][k] = v
        d["metadata"]["mesh_dx"] = mesh_dx

        return cls(d, origin, mesh_dx)

    def getLength(self):
        """
        Compute length of rupture (km). For EdgeRupture, we compute the length
        as the length of the top edge projected to the surface.

        Returns:
            float: Rupture length in km.
        """
        lons = self._toplons
        lats = self._toplats
        seg = self._group_index
        groups = np.unique(seg)
        ng = len(groups)
        rlength = 0
        for i in range(ng):
            group_segments = np.where(groups[i] == seg)[0]
            nseg = len(group_segments) - 1
            for j in range(nseg):
                ind = group_segments[j]
                P0 = Point(lons[ind], lats[ind])
                P1 = Point(lons[ind + 1], lats[ind + 1])
                dist = P0.distance(P1)
                rlength = rlength + dist
        return rlength

    def getWidth(self):
        """
        Compute average rupture width in km. For EdgeRupture, we compute this
        as the rupture area divided by teh rupture length.

        Returns:
            float: Rupture width in km.
        """
        area = self.getArea()
        length = self.getLength()
        return area / length

    def getArea(self):
        """
        Compute the rupture area. For EdgeRupture, we compute this by grouping
        the traces into "quadrilaterals" for which the verticies may not be
        co-planar. We then break up the quadrilaterals into triangles for which
        we can compute area.

        Returns:
            float: Rupture area in square km.
        """
        seg = self._group_index
        groups = np.unique(seg)
        ng = len(groups)
        area = 0
        for i in range(ng):
            group_segments = np.where(groups[i] == seg)[0]
            nseg = len(group_segments) - 1
            for j in range(nseg):
                ind = group_segments[j]
                p0 = latlon2ecef(
                    self._toplats[ind], self._toplons[ind], self._topdeps[ind]
                )
                p1 = latlon2ecef(
                    self._toplats[ind + 1],
                    self._toplons[ind + 1],
                    self._topdeps[ind + 1],
                )
                p2 = latlon2ecef(
                    self._botlats[ind + 1],
                    self._botlons[ind + 1],
                    self._botdeps[ind + 1],
                )
                p3 = latlon2ecef(
                    self._botlats[ind], self._botlons[ind], self._botdeps[ind]
                )
                a = np.sqrt(
                    (p1[0] - p0[0]) ** 2
                    + (p1[1] - p0[1]) ** 2
                    + (p1[2] - p0[2]) ** 2
                )
                b = np.sqrt(
                    (p2[0] - p0[0]) ** 2
                    + (p2[1] - p0[1]) ** 2
                    + (p2[2] - p0[2]) ** 2
                )
                c = np.sqrt(
                    (p2[0] - p1[0]) ** 2
                    + (p2[1] - p1[1]) ** 2
                    + (p2[2] - p1[2]) ** 2
                )
                s = (a + b + c) / 2
                A1 = np.sqrt(s * (s - a) * (s - b) * (s - c))
                a = np.sqrt(
                    (p0[0] - p3[0]) ** 2
                    + (p0[1] - p3[1]) ** 2
                    + (p0[2] - p3[2]) ** 2
                )
                b = np.sqrt(
                    (p2[0] - p3[0]) ** 2
                    + (p2[1] - p3[1]) ** 2
                    + (p2[2] - p3[2]) ** 2
                )
                c = np.sqrt(
                    (p0[0] - p2[0]) ** 2
                    + (p0[1] - p2[1]) ** 2
                    + (p0[2] - p2[2]) ** 2
                )
                s = (a + b + c) / 2
                A2 = np.sqrt(s * (s - a) * (s - b) * (s - c))
                area = area + (A1 + A2) / 1000 / 1000
        return area

    def getStrike(self):
        """
        Return representative strike for this rupture. Note that strike
        can vary along the rupture.

        Returns:
            float: Strike angle in degrees.
        """
        return self._strike

    def getDip(self):
        """
        Representative dip for this rupture. Note that dip
        can vary along the rupture.

        Returns:
            float: dip angle in degrees.
        """
        return self._dip

    @property
    def lats(self):
        """
        Return an array of latitudes for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop latitude values; disconnected
                segments are separated by nans.
        """
        lats = np.empty(shape=(0,))
        groups = self._group_index
        u_groups = np.unique(groups)
        ng = len(u_groups)
        nan = np.array(np.nan).reshape(
            1,
        )
        for i in range(ng):
            top_lats = self._toplats[groups == u_groups[i]]
            top0 = top_lats[0].reshape((1,))
            bot_lats = self._botlats[groups == u_groups[i]]
            lats = np.concatenate((lats, top_lats, bot_lats[::-1], top0, nan))
        return np.array(lats)

    @property
    def lons(self):
        """
        Return an array of longitudes for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop longitude values; disconnected
                segments are separated by nans.
        """
        lons = np.empty(shape=(0,))
        groups = self._group_index
        u_groups = np.unique(groups)
        ng = len(u_groups)
        nan = np.array(np.nan).reshape(
            1,
        )
        for i in range(ng):
            top_lons = self._toplons[groups == u_groups[i]]
            top0 = top_lons[0].reshape((1,))
            bot_lons = self._botlons[groups == u_groups[i]]
            lons = np.concatenate((lons, top_lons, bot_lons[::-1], top0, nan))
        return np.array(lons)

    @property
    def depths(self):
        """
        Return an array of depths for the rupture verticies arranged for
        plotting purposes; will give an outline of each group connected
        segments.

        Returns:
            array: Numpy array of closed-loop latitude values; disconnected
                segments are separated by nans.
        """
        deps = np.empty(shape=(0,))
        groups = self._group_index
        u_groups = np.unique(groups)
        ng = len(u_groups)
        nan = np.array(np.nan).reshape(
            1,
        )
        for i in range(ng):
            top_deps = self._topdeps[groups == u_groups[i]]
            top0 = top_deps[0].reshape((1,))
            bot_deps = self._botdeps[groups == u_groups[i]]
            deps = np.concatenate((deps, top_deps, bot_deps[::-1], top0, nan))
        return np.array(deps)

    def _getGroupIndex(self):
        # This is a bit hacky because the _group_index attribute is different
        # for an edge rupture than for quad rupture. Maybe re-doing the code
        # at a lower level would be a better resolution to this problem.
        ugroup = np.unique(self._group_index)
        ngroup = len(ugroup)
        fixed_group_index = []
        for i in range(ngroup):
            ind = np.where(self._group_index == ugroup[i])[0]
            nq = len(ind) - 1
            fixed_group_index.extend([i] * nq)
        return fixed_group_index

    def _computeStrikeDip(self):
        """
        Loop over all triangles and get the average normal, north, and up
        vectors in ECEF. Use these to compute a representative strike and dip.
        """
        seg = self._group_index
        groups = np.unique(seg)
        ng = len(groups)
        norm_vec = Vector(0, 0, 0)
        north_vec = Vector(0, 0, 0)
        up_vec = Vector(0, 0, 0)
        for i in range(ng):
            group_segments = np.where(groups[i] == seg)[0]
            nseg = len(group_segments) - 1
            for j in range(nseg):
                ind = group_segments[j]
                P0 = Point(
                    self._toplons[ind], self._toplats[ind], self._topdeps[ind]
                )
                P1 = Point(
                    self._toplons[ind + 1],
                    self._toplats[ind + 1],
                    self._topdeps[ind + 1],
                )
                P2 = Point(
                    self._botlons[ind + 1],
                    self._botlats[ind + 1],
                    self._botdeps[ind + 1],
                )
                P3 = Point(
                    self._botlons[ind], self._botlats[ind], self._botdeps[ind]
                )
                P1up = Point(
                    self._toplons[ind + 1],
                    self._toplats[ind + 1],
                    self._topdeps[ind + 1] - 1.0,
                )
                P1N = Point(
                    self._toplons[ind + 1],
                    self._toplats[ind + 1] + 0.001,
                    self._topdeps[ind + 1],
                )
                P3up = Point(
                    self._botlons[ind],
                    self._botlats[ind],
                    self._botdeps[ind] - 1.0,
                )
                P3N = Point(
                    self._botlons[ind],
                    self._botlats[ind] + 0.001,
                    self._botdeps[ind],
                )
                p0 = Vector.fromPoint(P0)
                p1 = Vector.fromPoint(P1)
                p2 = Vector.fromPoint(P2)
                p3 = Vector.fromPoint(P3)
                p1up = Vector.fromPoint(P1up)
                p1N = Vector.fromPoint(P1N)
                p3up = Vector.fromPoint(P3up)
                p3N = Vector.fromPoint(P3N)

                # Sides
                s01 = p1 - p0
                s02 = p2 - p0
                s03 = p3 - p0
                s21 = p1 - p2
                s23 = p3 - p2

                # First triangle
                t1norm = (s02.cross(s01)).norm()
                a = s01.mag()
                b = s02.mag()
                c = s21.mag()
                s = (a + b + c) / 2
                A1 = np.sqrt(s * (s - a) * (s - b) * (s - c)) / 1000

                # Second triangle
                t2norm = (s03.cross(s02)).norm()
                a = s03.mag()
                b = s23.mag()
                c = s02.mag()
                s = (a + b + c) / 2
                A2 = np.sqrt(s * (s - a) * (s - b) * (s - c)) / 1000

                # Up and North
                p1up = (p1up - p1).norm()
                p3up = (p3up - p3).norm()
                p1N = (p1N - p1).norm()
                p3N = (p3N - p3).norm()

                # Combine
                norm_vec = norm_vec + A1 * t1norm + A2 * t2norm
                north_vec = north_vec + A1 * p1N + A2 * p3N
                up_vec = up_vec + A1 * p1up + A2 * p3up

        norm_vec = norm_vec.norm()
        north_vec = north_vec.norm()
        up_vec = up_vec.norm()

        # Do I need to flip the vector because it is pointing down (i.e.,
        # right-hand rule is violated)?
        flip = np.sign(up_vec.dot(norm_vec))
        norm_vec = flip * norm_vec

        # Angle between up_vec and norm_vec is dip
        self._dip = np.arcsin(up_vec.cross(norm_vec).mag()) * 180 / np.pi

        # Normal vector projected to horizontal plane
        nvph = (norm_vec - up_vec.dot(norm_vec) * up_vec).norm()

        # Dip direction is angle between nvph and north; strike is orthogonal.
        cp = nvph.cross(north_vec)
        sign = np.sign(cp.dot(up_vec))
        dp = nvph.dot(north_vec)
        strike = np.arctan2(sign * cp.mag(), dp) * 180 / np.pi - 90
        if strike < -180:
            strike = strike + 360
        self._strike = strike

    def getDepthToTop(self):
        """
        Returns:
            float: Depth to top of rupture in km.
        """
        return np.min(self._topdeps)

    def getQuadrilaterals(self):
        """
        Return a list of quadrilaterals. Unlike QuadRupture, these
        quadrilaterals are not restricted to be coplanar or have
        horizontal top/bottom edges.

        Return:
            list: List of quadrilaterals, where each quadrilateral is
                a list of OQ Points.
        """
        ugroup = np.unique(self._group_index)
        ngroup = len(ugroup)
        qlist = []
        for i in range(ngroup):
            ind = np.where(self._group_index == ugroup[i])[0]
            nq = len(ind) - 1
            for j in range(nq):
                P0 = Point(
                    self._toplons[j], self._toplats[j], self._topdeps[j]
                )
                P1 = Point(
                    self._toplons[j + 1],
                    self._toplats[j + 1],
                    self._topdeps[j + 1],
                )
                P2 = Point(
                    self._botlons[j + 1],
                    self._botlats[j + 1],
                    self._botdeps[j + 1],
                )
                P3 = Point(
                    self._botlons[j], self._botlats[j], self._botdeps[j]
                )
                qlist.append([P0, P1, P2, P3])

        return qlist

    def computeRrup(self, lon, lat, depth):
        """
        Method for computing rupture distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           tuple: A tuple of an array of rupture distance (km), and None.

        """

        mesh_dx = self._mesh_dx

        # ---------------------------------------------------------------------
        # Sort out sites
        # ---------------------------------------------------------------------
        oldshape = lon.shape

        if len(oldshape) == 2:
            newshape = (oldshape[0] * oldshape[1], 1)
        else:
            newshape = (oldshape[0], 1)

        x, y, z = latlon2ecef(lat, lon, depth)
        x.shape = newshape
        y.shape = newshape
        z.shape = newshape
        sites_ecef = np.hstack((x, y, z))

        # ---------------------------------------------------------------------
        # Get mesh
        # ---------------------------------------------------------------------
        mx = []
        my = []
        mz = []
        u_groups = np.unique(self._group_index)
        n_groups = len(u_groups)
        for j in range(n_groups):
            g_ind = np.where(u_groups[j] == self._group_index)[0]
            nq = len(self._toplats[g_ind]) - 1
            for i in range(nq):
                q = [
                    Point(
                        self._toplons[g_ind[i]],
                        self._toplats[g_ind[i]],
                        self._topdeps[g_ind[i]],
                    ),
                    Point(
                        self._toplons[g_ind[i + 1]],
                        self._toplats[g_ind[i + 1]],
                        self._topdeps[g_ind[i + 1]],
                    ),
                    Point(
                        self._botlons[g_ind[i + 1]],
                        self._botlats[g_ind[i + 1]],
                        self._botdeps[g_ind[i + 1]],
                    ),
                    Point(
                        self._botlons[g_ind[i]],
                        self._botlats[g_ind[i]],
                        self._botdeps[g_ind[i]],
                    ),
                ]
                mesh = utils.get_quad_mesh(q, dx=mesh_dx)
                mx.extend(list(np.reshape(mesh["x"], (-1,))))
                my.extend(list(np.reshape(mesh["y"], (-1,))))
                mz.extend(list(np.reshape(mesh["z"], (-1,))))
        mesh_mat = np.array([np.array(mx), np.array(my), np.array(mz)])

        # ---------------------------------------------------------------------
        # Compute distance
        # ---------------------------------------------------------------------
        dist = np.zeros_like(x)
        for i in range(len(x)):
            sitecol = sites_ecef[i, :].reshape([3, 1])
            dif = sitecol - mesh_mat
            distarray = np.sqrt(np.sum(dif * dif, axis=0))
            dist[i] = np.min(distarray) / 1000.0  # convert to km

        dist = np.reshape(dist, oldshape)

        return dist

    def computeRjb(self, lon, lat, depth):
        """
        Method for computing Joyner-Boore distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           tuple: A tuple of an array of Joyner-Boore distance (km), and None.

        """

        mesh_dx = self._mesh_dx

        # ---------------------------------------------------------------------
        # Sort out sites
        # ---------------------------------------------------------------------
        oldshape = lon.shape

        if len(oldshape) == 2:
            newshape = (oldshape[0] * oldshape[1], 1)
        else:
            newshape = (oldshape[0], 1)

        x, y, z = latlon2ecef(lat, lon, depth)
        x.shape = newshape
        y.shape = newshape
        z.shape = newshape
        sites_ecef = np.hstack((x, y, z))

        # ---------------------------------------------------------------------
        # Get mesh
        # ---------------------------------------------------------------------
        mx = []
        my = []
        mz = []
        u_groups = np.unique(self._group_index)
        n_groups = len(u_groups)
        for j in range(n_groups):
            g_ind = np.where(u_groups[j] == self._group_index)[0]
            nq = len(self._toplats[g_ind]) - 1
            for i in range(nq):
                q = [
                    Point(self._toplons[g_ind[i]], self._toplats[g_ind[i]], 0),
                    Point(
                        self._toplons[g_ind[i + 1]],
                        self._toplats[g_ind[i + 1]],
                        0,
                    ),
                    Point(
                        self._botlons[g_ind[i + 1]],
                        self._botlats[g_ind[i + 1]],
                        0,
                    ),
                    Point(self._botlons[g_ind[i]], self._botlats[g_ind[i]], 0),
                ]
                mesh = utils.get_quad_mesh(q, dx=mesh_dx)
                mx.extend(list(np.reshape(mesh["x"], (-1,))))
                my.extend(list(np.reshape(mesh["y"], (-1,))))
                mz.extend(list(np.reshape(mesh["z"], (-1,))))
        mesh_mat = np.array([np.array(mx), np.array(my), np.array(mz)])

        # ---------------------------------------------------------------------
        # Compute distance
        # ---------------------------------------------------------------------
        dist = np.zeros_like(x)
        for i in range(len(x)):
            sitecol = sites_ecef[i, :].reshape([3, 1])
            dif = sitecol - mesh_mat
            distarray = np.sqrt(np.sum(dif * dif, axis=0))
            dist[i] = np.min(distarray) / 1000.0  # convert to km

        dist = np.reshape(dist, oldshape)

        return dist

    def computeGC2(self, lon, lat, depth):
        """
        Method for computing version 2 of the Generalized Coordinate system
        (GC2) by Spudich and Chiou OFR 2015-1028.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            dict: Dictionary with keys for each of the GC2-related distances,
                which include 'rx', 'ry', 'ry0', 'U', 'T'.
        """
        # This just hands off to the module-level method
        # NOTE: Not sure if the non-horizontal top edges of EdgeRupture will
        #       case problems. Should do some checking. It might be okay to
        #       bring quad vertices up to the surface in this case.
        dict = gc2._computeGC2(self, lon, lat, depth)
        return dict



#!/usr/bin/env python

from collections import OrderedDict
import numpy as np
from esi_utils_rupture.tensor import (
    fill_tensor_from_angles,
    fill_tensor_from_components,
    plane_to_tensor,
)
import datetime


def test_tensor():
    # from https://earthquake.usgs.gov/earthquakes/eventpage/us2000ar20#moment-tensor
    event = OrderedDict(
        [
            ("id", "us2000ar20"),
            ("time", datetime.datetime(2017, 9, 19, 18, 14, 38, 380000)),
            ("location", "1km ESE of Ayutla, Mexico"),
            ("latitude", 18.5462),
            ("longitude", -98.4871),
            ("depth", 51),
            ("magnitude", 7.1),
            ("magtype", "mww"),
            ("url", "https://earthquake.usgs.gov/earthquakes/eventpage/us2000ar20"),
            ("us_Mwb_mrr", -4.7728e19),
            ("us_Mwb_mtt", 3.3005e19),
            ("us_Mwb_mpp", 1.4723e19),
            ("us_Mwb_mrt", 1.481e18),
            ("us_Mwb_mrp", -4.319e18),
            ("us_Mwb_mtp", -3.305e19),
            ("us_Mwb_np1_strike", "123.35"),
            ("us_Mwb_np1_dip", "43.08"),
            ("us_Mwb_np1_rake", "-95.65"),
            ("us_Mwb_np2_strike", "311.07"),
            ("us_Mwb_np2_dip", "47.18"),
            ("us_Mwb_np2_rake", "-84.74"),
        ]
    )
    strike = float(event["us_Mwb_np1_strike"])
    dip = float(event["us_Mwb_np1_dip"])
    rake = float(event["us_Mwb_np1_rake"])
    mag = event["magnitude"]

    comps = plane_to_tensor(strike, dip, rake, mag=mag)
    mrr = comps[0][0]
    mtt = comps[1][1]
    mpp = comps[2][2]
    mrt = comps[0][1]
    mrp = comps[2][0]
    mtp = comps[1][2]
    rtol = 0.5  # how should this be chosen?
    atol = 0.0
    np.testing.assert_allclose(mrr, event["us_Mwb_mrr"], rtol=rtol, atol=atol)
    np.testing.assert_allclose(mtt, event["us_Mwb_mtt"], rtol=rtol, atol=atol)
    np.testing.assert_allclose(mpp, event["us_Mwb_mpp"], rtol=rtol, atol=atol)
    np.testing.assert_allclose(mrt, event["us_Mwb_mrt"], rtol=rtol, atol=atol)
    np.testing.assert_allclose(mrp, event["us_Mwb_mrp"], rtol=rtol, atol=atol)
    np.testing.assert_allclose(mtp, event["us_Mwb_mtp"], rtol=rtol, atol=atol)

    tensor_from_angles = fill_tensor_from_angles(strike, dip, rake, magnitude=mag)
    np.testing.assert_allclose(
        tensor_from_angles["mrr"], event["us_Mwb_mrr"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_angles["mtt"], event["us_Mwb_mtt"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_angles["mpp"], event["us_Mwb_mpp"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_angles["mrt"], event["us_Mwb_mrt"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_angles["mrp"], event["us_Mwb_mrp"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_angles["mtp"], event["us_Mwb_mtp"], rtol=rtol, atol=atol
    )

    try:
        np.testing.assert_allclose(
            strike, tensor_from_angles["NP1"]["strike"], rtol=0.0001, atol=0
        )
        np.testing.assert_allclose(
            dip, tensor_from_angles["NP1"]["dip"], rtol=0.0001, atol=0
        )
        np.testing.assert_allclose(
            rake, tensor_from_angles["NP1"]["rake"], rtol=0.0001, atol=0
        )
    except AssertionError as ae:
        try:
            np.testing.assert_allclose(
                strike, tensor_from_angles["NP2"]["strike"], rtol=0.0001, atol=0
            )
            np.testing.assert_allclose(
                dip, tensor_from_angles["NP2"]["dip"], rtol=0.0001, atol=0
            )
            np.testing.assert_allclose(
                rake, tensor_from_angles["NP2"]["rake"], rtol=0.0001, atol=0
            )
        except AssertionError as ae2:
            raise AssertionError("Nodal plane angles are off.")

    mrr = event["us_Mwb_mrr"]
    mtt = event["us_Mwb_mtt"]
    mpp = event["us_Mwb_mpp"]
    mrt = event["us_Mwb_mrt"]
    mrp = event["us_Mwb_mrp"]
    mtp = event["us_Mwb_mtp"]

    tensor_from_comps = fill_tensor_from_components(mrr, mtt, mpp, mrt, mrp, mtp)
    np.testing.assert_allclose(
        tensor_from_comps["mrr"], event["us_Mwb_mrr"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_comps["mtt"], event["us_Mwb_mtt"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_comps["mpp"], event["us_Mwb_mpp"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_comps["mrt"], event["us_Mwb_mrt"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_comps["mrp"], event["us_Mwb_mrp"], rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        tensor_from_comps["mtp"], event["us_Mwb_mtp"], rtol=rtol, atol=atol
    )

    try:
        np.testing.assert_allclose(
            strike, tensor_from_comps["NP1"]["strike"], rtol=0.0001, atol=0
        )
        np.testing.assert_allclose(
            dip, tensor_from_comps["NP1"]["dip"], rtol=0.0001, atol=0
        )
        np.testing.assert_allclose(
            rake, tensor_from_comps["NP1"]["rake"], rtol=0.0001, atol=0
        )
    except AssertionError as ae:
        try:
            np.testing.assert_allclose(
                strike, tensor_from_comps["NP2"]["strike"], rtol=0.0001, atol=0
            )
            np.testing.assert_allclose(
                dip, tensor_from_comps["NP2"]["dip"], rtol=0.0001, atol=0
            )
            np.testing.assert_allclose(
                rake, tensor_from_comps["NP2"]["rake"], rtol=0.0001, atol=0
            )
        except AssertionError as ae2:
            raise AssertionError("Nodal plane angles are off.")


if __name__ == "__main__":
    test_tensor()



#!/usr/bin/env python

# stdlib modules
from abc import ABC
from abc import abstractmethod
import json

# third party imports
import numpy as np

try:
    _ = np.RankWarning  # will work on numpy < 2
except AttributeError:
    setattr(np, "RankWarning", RuntimeWarning)  # will work on numpy > 2
from openquake.hazardlib.geo import geodetic

# local imports
from esi_utils_rupture import constants


class RuptureContext:
    pass


class Rupture(ABC):
    """
    Abstract base class for ruptures.

    Note on terminology:

        - There are three Ruptuer subclasses: PointRupture, QuadRupture, and
          EdgeRupture.
        - PointRupture represents the rupture as a point source.
        - QuadRupture and EdgeRupture are two different finite source
          representations.
        - A finite rupture is composed of segments. For QuadRupture, a segment
          is a quadrilaterial; for an EdgeRupture, a segment is a line
          connecting two points.
        - Segments are grouped with a common "group index".
        - Segments within a group must be continuous.
        - The QuadRupture class requires that each segment is a quadrilateral
          with horizontal top and obttom edges.
        - The EdgeRupture class allows for arbitrarily complex top and bottom
          edge specification.

    """

    def writeGeoJson(self, file):
        """
        Write the rupture to a GeoJson file.

        Args:
            file (str): Name of file.
        """
        with open(file, "w") as f:
            json.dump(self._geojson, f)

    def getGeoJson(self):
        """
        Get the object's GeoJson
        """
        return self._geojson

    @abstractmethod
    def getLength(self):  # pragma: no cover
        """
        Returns:
            float: Rupture length in km.
        """
        raise NotImplementedError(
            "'Rupture::getLength' must be overridden by child classes"
        )

    @abstractmethod
    def getWidth(self):  # pragma: no cover
        """
        Returns:
            float: Rupture width in km.
        """
        raise NotImplementedError(
            "'Rupture::getWidth' must be overridden by child classes"
        )

    @abstractmethod
    def getArea(self):  # pragma: no cover
        """
        Returns:
            float: Rupture area in square km.
        """
        raise NotImplementedError(
            "'Rupture::getArea' must be overridden by child classes"
        )

    @abstractmethod
    def getStrike(self):  # pragma: no cover
        """
        Return strike angle. If rupture consists of multiple quadrilaterals,
        the average strike angle, weighted by quad length, is returned.
        Note: for ruptures with quads where the strike angle changes by 180 deg
        due to reverses in dip direction are problematic and not handeled well
        by this algorithm.

        Returns:
            float: Strike angle in degrees.

        """
        raise NotImplementedError(
            "'Rupture::getStrike' must be overridden by child classes"
        )

    @abstractmethod
    def getDip(self):  # pragma: no cover
        raise NotImplementedError(
            "'Rupture::getDip' must be overridden by child classes"
        )

    @abstractmethod
    def getDepthToTop(self):  # pragma: no cover
        """
        Returns:
           float: Average dip in degrees.

        """
        raise NotImplementedError(
            "'Rupture::getDepthToTop' must be overridden by child classes"
        )

    @abstractmethod
    def getQuadrilaterals(self):  # pragma: no cover
        """
        Method to return rupture quadrilaterals. Returns None for
        PointRupture.
        """
        raise NotImplementedError(
            "'Rupture::getQuadrilaterals' must be overridden by child classes"
        )

    def getReference(self):
        """
        Returns:
           string: Reference info from file.

        """
        return self._reference

    def getOrigin(self):
        """
        Returns:
           Origin object

        """
        return self._origin

    @property
    @abstractmethod
    def lats(self):  # pragma: no cover
        raise NotImplementedError(
            "'Rupture::lats' must be overridden by child classes"
        )

    @property
    @abstractmethod
    def lons(self):  # pragma: no cover
        raise NotImplementedError(
            "'Rupture::lons' must be overridden by child classes"
        )

    @property
    @abstractmethod
    def depths(self):  # pragma: no cover
        raise NotImplementedError(
            "'Rupture::depths' must be overridden by child classes"
        )

    def getRuptureContext(self, gmpelist, shape=None):
        """
        Args:
            gmpelist (list): List of hazardlib GMPE objects.
            shape (tuple): The shape of the ndarrays for each element. If "None",
                the values will be scalars.


        Returns:
            RuptureContext object with all known parameters filled in.

        """  # noqa

        origin = self._origin

        # rupturecontext constructor inputs:
        # 'mag', 'strike', 'dip', 'rake', 'ztor', 'hypo_lon', 'hypo_lat',
        # 'hypo_depth', 'width', 'hypo_loc'

        rx = RuptureContext()
        if shape is None:
            rx.mag = origin.mag
            rx.strike = self.getStrike()
            rx.dip = self.getDip()
            rx.ztor = self.getDepthToTop()
            rx.width = self.getWidth()

            if hasattr(origin, "rake"):
                rx.rake = origin.rake
            elif hasattr(origin, "mech"):
                rx.rake = constants.RAKEDICT[origin.mech]
            else:
                rx.rake = constants.RAKEDICT["ALL"]

            rx.hypo_lat = origin.lat
            rx.hypo_lon = origin.lon
            rx.hypo_depth = origin.depth
        else:
            rx.mag = np.full(shape, origin.mag)
            rx.strike = np.full(shape, self.getStrike())
            rx.dip = np.full(shape, self.getDip())
            rx.ztor = np.full(shape, self.getDepthToTop())
            rx.width = np.full(shape, self.getWidth())

            if hasattr(origin, "rake"):
                rx.rake = np.full(shape, origin.rake)
            elif hasattr(origin, "mech"):
                rx.rake = np.full(shape, constants.RAKEDICT[origin.mech])
            else:
                rx.rake = np.full(shape, constants.RAKEDICT["ALL"])

            rx.hypo_lat = np.full(shape, origin.lat)
            rx.hypo_lon = np.full(shape, origin.lon)
            rx.hypo_depth = np.full(shape, origin.depth)

        return rx

    def computeRhyp(self, lon, lat, depth):
        """
        Method for computing hypocentral distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           array: Hypocentral distance (km).
        """
        origin = self._origin
        oldshape = lon.shape

        rhyp = geodetic.distance(
            origin.lon, origin.lat, origin.depth, lon, lat, depth
        )
        rhyp = rhyp.reshape(oldshape)
        return rhyp

    def computeRepi(self, lon, lat, depth):
        """
        Method for computing epicentral distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           array: Epicentral distance (km).
        """
        origin = self._origin
        oldshape = lon.shape

        repi = geodetic.distance(origin.lon, origin.lat, 0.0, lon, lat, depth)
        repi = repi.reshape(oldshape)
        return repi

    @abstractmethod
    def computeRjb(self, lon, lat, depth):  # pragma: no cover
        """
        Method for computing Joyner-Boore distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           array: Joyner-Boore distance (km).

        """
        pass

    @abstractmethod
    def computeRrup(self, lon, lat, depth):  # pragma: no cover
        """
        Method for computing rupture distance.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
           array: Rupture distance (km).

        """
        pass

    @abstractmethod
    def computeGC2(self, lon, lat, depth):  # pragma: no cover
        """
        Method for computing version 2 of the Generalized Coordinate system
        (GC2) by Spudich and Chiou OFR 2015-1028.

        Args:
            lon (array): Numpy array of longitudes.
            lat (array): Numpy array of latitudes.
            depth (array): Numpy array of depths (km; positive down).

        Returns:
            dict: Dictionary with keys for each of the GC2-related distances,
                which include 'rx', 'ry', 'ry0', 'U', 'T'.
        """
        pass
